import multiprocessing as mp
import time
from pathlib import Path
from threading import Lock, Thread
from typing import Callable

from apscheduler.schedulers.background import BackgroundScheduler

from ldnode.runner.config import Settings
from ldnode.runner.schemas import Worker

from .datasources import (
    CBQuotationsDataSource,
    COSDataSource,
    ETFQuotationsDataSource,
    S3QuotationDataSource,
    StkQuotationsDataSource,
)
from .files import (
    CB_DATA,
    CB_HIST_DATA,
    CB_PRE_QUOTATION,
    CB_QUOTATION,
    ETF_DATA,
    ETF_PRE_QUOTATION,
    ETF_QUOTATION,
    INDEX_DATA,
    STK_DATA,
    STK_HIST_DATA,
    STK_PRE_QUOTATION,
    STK_QUOTATION,
    TRADE_DAYS,
)
from .persistence import LightCOSClient
from .schemas.models import LoadRule, LoadTrigger, SyncMetadata
from .stop import is_trading_time


class ContinuousLoader:
    """连续加载器"""

    def __init__(self):
        self.threads: dict[str, Thread] = {}

    def _cycle(self, func: Callable[[], bool]) -> None:
        continue_flag = True
        while continue_flag:
            continue_flag = func()

    def add_job(self, key: str, func: Callable[[], bool]) -> None:
        if key not in self.threads:
            self.threads[key] = Thread(target=self._cycle, args=(func,), daemon=True)

    def start(self) -> None:
        for thread in self.threads.values():
            thread.start()

    def shutdown(self) -> None:
        for thread in self.threads.values():
            thread.join()


class LoaderWorker(Worker):
    """加载工作进程"""

    def __init__(self, worker_name: str):
        super().__init__(worker_name)
        self.ap_scheduler = BackgroundScheduler()
        self.continuous_loader = ContinuousLoader()
        self.locks: dict[str, Lock] = {}
        self.output: mp.Queue | None = None

    def _get_cos_datasource(
        self, settings: Settings, metadata: SyncMetadata
    ) -> COSDataSource:
        cos_client = LightCOSClient(
            region=settings.DATA_SYNC_COS_REGION,
            secret_id=settings.DATA_SYNC_COS_SECRET_ID,
            secret_key=settings.DATA_SYNC_COS_SECRET_KEY,
            bucket=settings.DATA_SYNC_COS_BUCKET,
        )
        return COSDataSource(
            cos_client, metadata, Path(settings.DATA_SYNC_LOCAL_DATA_DIR)
        )

    def _wrap_load(self, rule: LoadRule) -> Callable[[], bool]:
        """包装加载函数"""

        def load() -> bool:
            if not self.running:
                return False
            if rule.is_stop and rule.is_stop():
                return True
            if rule.file.key not in self.locks:
                self.locks[rule.file.key] = Lock()
            with self.locks[rule.file.key]:
                results = rule.datasource.load(rule.file.key)
                if self.output:
                    for result in results:
                        self.output.put(result)
            return True

        return load

    def _get_s3_storage_options(self, settings: Settings) -> dict:
        """构造 S3 storage_options（包含 MinIO 兼容所需的 allow_http / region）"""
        opts: dict = {
            "allow_http": "true",
            "aws_region": "us-east-1",
        }
        if settings.QUOTATION_S3_KEY_ID:
            opts["aws_access_key_id"] = settings.QUOTATION_S3_KEY_ID
            opts["aws_secret_access_key"] = settings.QUOTATION_S3_SECRET_KEY
        if settings.QUOTATION_S3_ENDPOINT_URL:
            opts["endpoint_url"] = settings.QUOTATION_S3_ENDPOINT_URL
        return opts

    def init_load_rules(self, settings: Settings) -> list[LoadRule]:
        """初始化加载规则。
        配置了 S3 URI 的行情品种：注册 S3QuotationDataSource（后台每 3s 拉取写本地）。
        未配置 S3 URI：原有本地行情计算逻辑不变。
        output_lazyframe() 始终读本地文件，无感知切换。
        """
        local_data_dir = Path(settings.DATA_SYNC_LOCAL_DATA_DIR)
        metadata = SyncMetadata(local_data_dir)
        cos = self._get_cos_datasource(settings, metadata)
        s3_opts = self._get_s3_storage_options(settings)

        rules = [
            LoadRule(TRADE_DAYS, LoadTrigger.INTERVAL, cos, 10),
            LoadRule(INDEX_DATA, LoadTrigger.INTERVAL, cos, 10),
            LoadRule(CB_DATA, LoadTrigger.INTERVAL, cos, 10),
            LoadRule(CB_PRE_QUOTATION, LoadTrigger.INTERVAL, cos, 10),
            LoadRule(CB_HIST_DATA, LoadTrigger.INTERVAL, cos, 10),
        ]
        # 转债实时行情：S3 后台预拉取 OR 本地计算
        if settings.CB_QUOTATION_S3_URI:
            rules.append(LoadRule(
                CB_QUOTATION, LoadTrigger.INTERVAL,
                S3QuotationDataSource(settings.CB_QUOTATION_S3_URI, s3_opts, local_data_dir),
                3, is_stop=lambda: not is_trading_time(),
            ))
        else:
            rules.append(LoadRule(
                CB_QUOTATION, LoadTrigger.INTERVAL, CBQuotationsDataSource(), 3,
                is_stop=lambda: not is_trading_time(),
            ))

        rules += [
            LoadRule(STK_DATA, LoadTrigger.INTERVAL, cos, 10),
            LoadRule(STK_PRE_QUOTATION, LoadTrigger.INTERVAL, cos, 10),
            LoadRule(STK_HIST_DATA, LoadTrigger.INTERVAL, cos, 10),
        ]
        # 股票实时行情：S3 后台预拉取 OR 本地计算
        if settings.STK_QUOTATION_S3_URI:
            rules.append(LoadRule(
                STK_QUOTATION, LoadTrigger.INTERVAL,
                S3QuotationDataSource(settings.STK_QUOTATION_S3_URI, s3_opts, local_data_dir),
                3, is_stop=lambda: not is_trading_time(),
            ))
        else:
            rules.append(LoadRule(
                STK_QUOTATION, LoadTrigger.INTERVAL, StkQuotationsDataSource(), 3,
                is_stop=lambda: not is_trading_time(),
            ))

        rules += [
            LoadRule(ETF_DATA, LoadTrigger.INTERVAL, cos, 10),
            LoadRule(ETF_PRE_QUOTATION, LoadTrigger.INTERVAL, cos, 10),
        ]
        # ETF 实时行情：S3 后台预拉取 OR 本地计算
        if settings.ETF_QUOTATION_S3_URI:
            rules.append(LoadRule(
                ETF_QUOTATION, LoadTrigger.INTERVAL,
                S3QuotationDataSource(settings.ETF_QUOTATION_S3_URI, s3_opts, local_data_dir),
                3, is_stop=lambda: not is_trading_time(),
            ))
        else:
            rules.append(LoadRule(
                ETF_QUOTATION, LoadTrigger.INTERVAL, ETFQuotationsDataSource(), 3,
                is_stop=lambda: not is_trading_time(),
            ))

        return rules

    def setup_load_rules(self, settings: Settings) -> None:
        """设置加载规则"""
        for rule in self.init_load_rules(settings):
            match rule.tigger:
                case LoadTrigger.INTERVAL:
                    self.ap_scheduler.add_job(
                        self._wrap_load(rule),
                        "interval",
                        id=rule.file.key,
                        name=rule.file.key,
                        seconds=rule.interval,
                    )
                case LoadTrigger.CONTINUOUS:
                    self.continuous_loader.add_job(rule.file.key, self._wrap_load(rule))

    def run(self, settings: Settings, input: mp.Queue, output: mp.Queue) -> None:
        """运行工作进程"""
        self.output = output
        self.setup_load_rules(settings)
        self.ap_scheduler.start()
        self.continuous_loader.start()
        while self.running:
            time.sleep(1)
        self.ap_scheduler.shutdown()
        self.continuous_loader.shutdown()



import threading as th
import multiprocessing as mp
import queue
from typing import Optional

from ldnode import logger
from ldnode.runner.events import TaskSubmitted
from .models import (
    InputMessageType, 
    InputMessage, 
    InvokeResult,
)
from .worker_context import WorkerContext


class RecvController(th.Thread):
    """ 接收控制器 """
    def __init__(self, context: WorkerContext, input_channel: mp.Queue):
        super().__init__(daemon=True)
        self.context = context
        self.input_channel = input_channel
        
    def exit_if_main_notify(self, msg: Optional[InputMessage]) -> bool:
        """ 如果收到主进程通知，则退出 """
        if msg is None:
            self.context.exit()
            return True
        else:
            return False
        
    def run(self) -> None:
        """ 运行控制器 """
        while self.context.running:
            try:
                msg: Optional[InputMessage] = self.input_channel.get(block=True, timeout=1)
                if self.exit_if_main_notify(msg):
                    break
                match msg['msg_type']:
                    case InputMessageType.TASK_SUBMITTED:
                        task_submitted: TaskSubmitted = msg['payload']
                        self.context.load_task(task_submitted)
                    case InputMessageType.TASK_CANCELLED:
                        self.context.cancel_task()
                    case InputMessageType.INVOKE_ANSWER:
                        result: InvokeResult = msg['payload']
                        self.context.submit_invoke_answer(result['task_id'], result['result'])
            except queue.Empty:
                continue
            except Exception as e:
                logger.calculator.error(f"接收消息失败: {e}")



import polars as pl

from ldnode.calculator.schemas import HoldRangeType


def planning_mod_rank(
    frame: pl.LazyFrame | pl.DataFrame,
    threshold: int,
    hold_range_type: HoldRangeType,
    position_lf: pl.LazyFrame
) -> pl.LazyFrame:
    """ 规划修正排名 """
    lf = frame.lazy() if isinstance(frame, pl.DataFrame) else frame
    lf = (
        # 合并持仓中的volume
        lf.join(position_lf.select('code', 'volume'), on='code', how='left')
            .with_columns(
                pl.col('volume').fill_null(.0).alias('volume')
            )
            # 记录原排名
            .rename({'rank': 'ori_rank'})
    )
    # 计算修正排名表达式
    mod_rank_expr = (
        pl.when(pl.col('volume') > 0)
        .then(pl.col('ori_rank') - threshold)
        .otherwise(pl.col('ori_rank'))
        .alias('mod_rank')
    )
    # 按数量排名时
    match hold_range_type:
        case HoldRangeType.NUM:
            lf = (
                lf.with_columns(mod_rank_expr)
                    .sort(['mod_rank','volume'], descending=[False,True])
                    .with_columns(pl.col('mod_rank').rank(method='ordinal').alias('rank'))
            )
        case HoldRangeType.PCT:
            lf = lf.with_columns(mod_rank_expr.alias('rank'))
    return lf



from .cos import COSDataSource
from .cb_quotations import CBQuotationsDataSource
from .stk_quotations import StkQuotationsDataSource
from .etf_quotations import ETFQuotationsDataSource
from .s3_quotation import S3QuotationDataSource


__all__ = ["COSDataSource", "CBQuotationsDataSource", "StkQuotationsDataSource", "ETFQuotationsDataSource", "S3QuotationDataSource"]



from typing import Callable, Any
from enum import StrEnum, IntEnum

import polars as pl


class FactorType(IntEnum):
    """ 因子类型 """
    BUILDIN = 0 # 内置
    COMPUTED = 1 # 计算


class ExprType(IntEnum):
    """ 表达式类型 """
    SQL = 0 # sql表达式
    POLARS = 1 # polars表达式
    DSL = 2 # dsl表达式


class RankScope(IntEnum):
    """ 打分范围 """
    ALL = 0 # 全部范围
    SW_INDUSTRY_1 = 1 # 申万一级行业
    SW_INDUSTRY_2 = 2 # 申万二级行业
    SW_INDUSTRY_3 = 3 # 申万三级行业


class ExcludeOperator(StrEnum):
    """ 排除因子比较操作 """
    compare: Callable[[str, Any], pl.Expr]
    
    def __new__(cls, value: str, compare: Callable[[str, Any], pl.Expr]) -> 'ExcludeOperator':
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.compare = compare
        return obj

    # strict=False：String 列中非数值字符串转为 null，null 比较结果为 null（视为不满足条件），
    # 兼容 is_call 等原本为数值但被 cast 为 String 存储的因子
    GE = ('ge', lambda factor, value: pl.col(factor).cast(pl.Float64, strict=False) >= value) # 大于等于
    GT = ('gt', lambda factor, value: pl.col(factor).cast(pl.Float64, strict=False) > value) # 大于
    LE = ('le', lambda factor, value: pl.col(factor).cast(pl.Float64, strict=False) <= value) # 小于等于
    LT = ('lt', lambda factor, value: pl.col(factor).cast(pl.Float64, strict=False) < value) # 小于
    NE = ('ne', lambda factor, value: pl.col(factor).ne_missing(value)) # 不等于
    EQ = ('eq', lambda factor, value: pl.col(factor) == value) # 等于
    ISNULL = ('isnull', lambda factor, _: pl.col(factor).is_null()) # 空值
    GT_PCT = ('gt_pct', lambda factor, value: pl.col(f'{factor}_pct') > value) # 百分比大于
    LT_PCT = ('lt_pct', lambda factor, value: pl.col(f'{factor}_pct') < value) # 百分比小于
    CONTAINS = ('contains', lambda factor, value: pl.col(factor).str.contains(value)) # 包含
    NOT_CONTAINS = ('not_contains', lambda factor, value: ~pl.col(factor).str.contains(value)) # 不包含
    STARTS_WITH = ('starts_with', lambda factor, value: pl.col(factor).str.starts_with(value)) # 以...开头
    ENDS_WITH = ('ends_with', lambda factor, value: pl.col(factor).str.ends_with(value)) # 以...结尾
    NOT_STARTS_WITH = ('not_starts_with', lambda factor, value: ~pl.col(factor).str.starts_with(value)) # 不以...开头
    NOT_ENDS_WITH = ('not_ends_with', lambda factor, value: ~pl.col(factor).str.ends_with(value)) # 不以...结尾


class ProcessMissingDataMethod(StrEnum):
    """ 缺失值处理方式 """
    process: Callable[[str], pl.Expr]
    
    def __new__(cls, value: str, process: Callable[[str], pl.Expr]) -> 'ProcessMissingDataMethod':
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.process = process
        return obj


class NeutralizationMethod(StrEnum):
    """ 中性化方式 """
    process: Callable[[list[str]], list[pl.Expr]]
    is_discrete: bool
    
    def __new__(cls, value: str, process: Callable[[list[str]], list[pl.Expr]], is_discrete: bool) -> 'NeutralizationMethod':
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.process = process
        obj.is_discrete = is_discrete
        return obj



import asyncio

from ldnode import logger
from ldnode.runner.config import Settings
from ldnode.runner.schemas import Service, BaseContext
from ldnode.runner.task_manager import TaskManager
from ldnode.runner.events import (
    on, 
    Events, 
    TaskSubmitted,
    TaskFinished,
    TaskCanceled,
    TaskCancel,
    WorkerStateChanged,
)
from .application import (
    CPUScheduleController,
    ResourceController,
    LoadCounter,
)
from .handlers import handler_registry
from .worker import CalculatorWorker


class CalculatorService(Service):
    """ 计算服务 """
    def __init__(self, settings: Settings, context: BaseContext):
        super().__init__(settings, context)
        self.running = False
        self.tasks = TaskManager()
        self.exit_flag = asyncio.Event()
        self.waiting_tasks: list[TaskSubmitted] = []
        self.schedule_controller = CPUScheduleController(
            self.context, 
            cpu_resource=self.settings.CPU_RESOURCE, 
            executor_cls=CalculatorWorker, 
            running_checker=lambda: self.running,
        )
        self.resource_controller = ResourceController(
            self.context,
            cpu_resource=self.settings.CPU_RESOURCE, 
            gpu_resource=0,
        )
        self.load_counter = LoadCounter(
            self.context,
            load_querier=lambda: self.schedule_controller.load,
            running_checker=lambda: self.running,
        )

    async def start(self) -> None:
        """ 启动服务 """
        self.running = True
        # 启动调度器
        await self.schedule_controller.schedule()
        # 启动负载计算
        await self.load_counter.count()
        # 等待退出
        await self.exit_flag.wait()
    
    async def shutdown(self) -> None:
        """ 关闭服务 """
        self.running = False
        self.exit_flag.set()
        
    async def release_next_task(self) -> None:
        """ 释放下一个等待中的任务 """
        if self.waiting_tasks:
            next_task = self.waiting_tasks.pop(0)
            self.tasks.create_task(self.handle_task_submitted("calculator", next_task))
        
    @on(Events.WORKER_START)
    async def handle_worker_start(self, _: str, data: WorkerStateChanged) -> None:
        """ 处理工作进程启动 """
        await self.schedule_controller.awake_executor(data['worker_id'])

    @on(Events.WORKER_EXIT)
    async def handle_worker_exit(self, _: str, data: WorkerStateChanged) -> None:
        """ 处理工作进程退出 """
        await self.schedule_controller.sleep_executor(data['worker_id'])
        
    @on(Events.TASK_SUBMITTED)
    async def handle_task_submitted(self, _: str, task: TaskSubmitted) -> None:
        """ 处理任务提交 """
        if handler := handler_registry.get(task['method']):
            if await self.resource_controller.acquire(handler.cpu_usage, handler.gpu_usage):
                # 获取资源成功, 提交任务
                await self.schedule_controller.submit(task)
            else:
                # 进入队列
                self.waiting_tasks.append(task)
        else:
            logger.calculator.warning(f"未找到处理器: {task['method']}")

    @on(Events.TASK_FINISHED)
    async def handle_task_finished(self, source: str, data: TaskFinished) -> None:
        """ 处理任务完成 """
        logger.calculator.info(f"{source} 完成任务: {data['method']}({data['task_id']})")
        if handler := handler_registry.get(data['method']):
            # 释放资源
            await self.resource_controller.release(handler.cpu_usage, handler.gpu_usage)
            # 释放下一个等待中的任务
            await self.release_next_task()
            
    @on(Events.TASK_CANCEL)
    async def handle_task_cancel(self, _: str, data: TaskCancel) -> None:
        """ 处理任务取消 """
        logger.calculator.info(f"取消任务: {data['method']}({data['task_id']})")
        # 判断任务是否还在等待队列中
        if task := next(filter(lambda task: task['task_id'] == data['task_id'], self.waiting_tasks), None):
            # 取消任务
            self.waiting_tasks.remove(task)
            await self.context.submit_event('calculator', Events.TASK_CANCELED, {
                "task_id": data['task_id'],
                "method": data['method'],
                "is_released": False,
            })
        else:
            # 任务不在等待队列中
            await self.schedule_controller.cancel_task(data['task_id'])
            
    @on(Events.TASK_CANCELED)
    async def handle_task_canceled(self, _: str, data: TaskCanceled) -> None:
        """ 处理任务取消 """
        logger.calculator.info(f"任务取消: {data['method']}({data['task_id']})")
        if data['is_released']:
            if handler := handler_registry.get(data['method']):
            # 释放资源
                await self.resource_controller.release(handler.cpu_usage, handler.gpu_usage)
            else:
                logger.calculator.warning(f"未找到处理器: {data['method']}")



from typing import TypedDict, TypeVar, Generic

from ldnode.calculator.schemas import (
    BacktestParameters,
    CBBacktestParameters,
    STKBacktestParameters,
    ETFBacktestParameters,
)
from .common import TaskBaseParameters


BACKTEST_MODEL = TypeVar('BACKTEST_MODEL', bound=BacktestParameters)
class BaseBinningParameters(TaskBaseParameters, Generic[BACKTEST_MODEL]):
    """ 分箱回测参数 """
    strategy_config: BACKTEST_MODEL # 策略配置
    bins: int # 分箱数


class CBBinningParameters(BaseBinningParameters[CBBacktestParameters]):
    """ 转债分箱回测参数 """


class STKBinningParameters(BaseBinningParameters[STKBacktestParameters]):
    """ 股票分箱回测参数 """
    

class ETFBinningParameters(BaseBinningParameters[ETFBacktestParameters]):
    """ 基金分箱回测参数 """


class BinningResponse(TypedDict):
    """ 转债分箱回测响应 """
    ic_mean: float # 相关性均值
    ir: float # 信息比率
    return_df: bytes # 收益率曲线
    corr_df: bytes # 相关性


from easyquotation.sina import Sina


class SinaWithTimeoutPatch(Sina):
    """ 新浪数据源带超时时间补丁 """
    
    def get_stocks_by_range(self, params):
        """ Hook 获取股票数据 """
        headers = self._get_headers()
        r = self._session.get(self.stock_api + params, headers=headers, timeout=5)
        return r.text



from datetime import date, datetime

import numpy as np
import polars as pl

from ldnode.calculator.schemas import EPSILON, EvaluateData, GlanceEvaluateData

PERIODS = 242


def evaluate(
    frame: pl.LazyFrame | pl.DataFrame,
    columns_dict: dict[str, str] = {
        "equity": "equity",
        "time_return": "time_return",
        "drawdown": "drawdown",
        "turnover": "turnover",
    },
) -> EvaluateData:
    """评价回测结果"""
    equity_field = columns_dict.get("equity", "equity")
    time_return_field = columns_dict.get("time_return", "time_return")
    drawdown_field = columns_dict.get("drawdown", "drawdown")
    turnover_field = columns_dict.get("turnover", None)
    res = frame.lazy() if isinstance(frame, pl.DataFrame) else frame
    exprs = [
        pl.col(equity_field).last().alias("cum_asset"),  # 累计资产
        pl.col(equity_field).first().alias("first_equity"),  # 首日资产
        pl.col(time_return_field).first().alias("first_return"),  # 首日回报
        pl.len().alias("trading_days"),  # 交易天数
        pl.col(time_return_field)
        .mean()
        .replace(0, EPSILON)
        .alias("daily_pnl_mean"),  # 日盈亏均值
        pl.col(time_return_field)
        .std(ddof=1)
        .replace(0, EPSILON)
        .fill_null(EPSILON)
        .alias("daily_pnl_std"),  # 日盈亏标准差
        pl.col(time_return_field)
        .max()
        .replace(0, EPSILON)
        .alias("daily_max_profit"),  # 日最大盈利
        pl.col(time_return_field)
        .min()
        .replace(0, EPSILON)
        .alias("daily_max_loss"),  # 日最大亏损
        pl.col(drawdown_field)
        .min()
        .replace(0, EPSILON)
        .alias("daily_max_drawdown"),  # 日最大回撤
        (pl.col(time_return_field) > 0)
        .sum()
        .fill_null(EPSILON)
        .alias("profit_count"),  # 盈利次数
        pl.col(time_return_field)
        .filter(pl.col(time_return_field) > 0)
        .mean()
        .fill_null(EPSILON)
        .alias("daily_profit_mean"),  # 日盈利均值
        pl.col(time_return_field)
        .filter(pl.col(time_return_field) < 0)
        .mean()
        .fill_null(EPSILON)
        .alias("daily_loss_mean"),  # 日亏损均值
        pl.col(time_return_field)
        .filter(pl.col(time_return_field) < 0)
        .std(ddof=1)
        .fill_null(EPSILON)
        .alias("daily_loss_std"),  # 日亏损标准差
    ]
    if turnover_field:
        exprs.append(
            pl.col(turnover_field)
            .sum()
            .truediv(pl.len())
            .fill_null(EPSILON)
            .alias("daily_turnover"),  # 日均换手率
        )
    # 计算基础统计量
    stats = res.select(exprs).collect()
    # 计算衍生指标
    capital: float = round(
        stats["first_equity"].item() / (stats["first_return"].item() + 1), 2
    )  # 初始资金
    total_return: float = stats["cum_asset"].item() / capital - 1  # 总收益率
    annu_return: float = (total_return + 1) ** (
        PERIODS / stats["trading_days"].item()
    ) - 1  # 年化收益率
    daily_win_rate: float = (
        stats["profit_count"].item() / stats["trading_days"].item()
    )  # 日胜率

    daily_profit_loss_rate: float = (
        -stats["daily_profit_mean"].item() / stats["daily_loss_mean"].item()
    )  # 日盈亏比
    sharpe_ratio: float = (
        stats["daily_pnl_mean"].item() / stats["daily_pnl_std"].item()
    ) * np.sqrt(PERIODS)  # 夏普比率
    sortino_ratio: float = (
        stats["daily_pnl_mean"].item() / stats["daily_loss_std"].item()
    ) * np.sqrt(PERIODS)  # 索提诺比率
    calmar_ratio: float = (
        annu_return / -stats["daily_max_drawdown"].item()
    )  # 卡玛比比率

    daily_max_drawdown = stats["daily_max_drawdown"].item()  # 最大回撤
    daily_max_drawdown_end_index = (
        res.select(pl.col(drawdown_field).arg_min()).collect().item()
    )  # 最大回撤结束日期索引(如果有多处最大回撤，选最先出现的索引)
    _zero_drawdown_index_df = res.select(
        pl.arg_where(pl.col(drawdown_field) == 0).alias("zero_drawdown_index")
    ).collect()  # 最大回撤为0的日期索引df
    daily_max_drawdown_start_index = _zero_drawdown_index_df.select(
        pl.col("zero_drawdown_index")
        .filter(pl.col("zero_drawdown_index") < daily_max_drawdown_end_index)
        .last()
        .fill_null(0)
    ).item()  # 最大回撤开始日期索引
    daily_max_drawdown_duration_trade_days: int = (
        daily_max_drawdown_end_index - daily_max_drawdown_start_index
    )  # 最大回撤持续交易日
    daily_max_drawdown_start_date_lf = res.select(
        pl.col("trade_date").gather(daily_max_drawdown_start_index)
    )  # 最大回撤开始日期
    daily_max_drawdown_end_date_lf = res.select(
        pl.col("trade_date").gather(daily_max_drawdown_end_index)
    )  # 最大回撤结束日期
    daily_max_drawdown_start_date_df, daily_max_drawdown_end_date_df = pl.collect_all(
        [daily_max_drawdown_start_date_lf, daily_max_drawdown_end_date_lf]
    )
    daily_max_drawdown_start_date = (
        daily_max_drawdown_start_date_df.item()
    )  # 最大回撤开始日期
    daily_max_drawdown_end_date = (
        daily_max_drawdown_end_date_df.item()
    )  # 最大回撤结束日期
    daily_max_drawdown_duration_calendar_days = (
        daily_max_drawdown_end_date - daily_max_drawdown_start_date
    ).days  # 最大回撤持续日历天数
    evaluate_dict: EvaluateData = {
        "cum_asset": stats["cum_asset"].item(),
        "total_return": total_return,
        "trading_days": stats["trading_days"].item(),
        "annu_return": annu_return,
        "daily_turnover": stats["daily_turnover"].item() if turnover_field else None,
        "profit_count": stats["profit_count"].item(),
        "daily_win_rate": daily_win_rate,
        "daily_profit_mean": stats["daily_profit_mean"].item(),
        "daily_loss_mean": stats["daily_loss_mean"].item(),
        "daily_profit_loss_rate": daily_profit_loss_rate,
        "daily_pnl_mean": stats["daily_pnl_mean"].item(),
        "daily_pnl_std": stats["daily_pnl_std"].item(),
        "daily_loss_std": stats["daily_loss_std"].item(),
        "daily_max_profit": stats["daily_max_profit"].item(),
        "daily_max_loss": stats["daily_max_loss"].item(),
        "daily_max_drawdown": daily_max_drawdown,
        "daily_max_drawdown_duration_trade_days": daily_max_drawdown_duration_trade_days,
        "daily_max_drawdown_duration_calendar_days": daily_max_drawdown_duration_calendar_days,
        "sharpe_ratio": sharpe_ratio,
        "calmar_ratio": calmar_ratio,
        "sortino_ratio": sortino_ratio,
    }
    return {
        k: round(v, 4) if isinstance(v, float) else v for k, v in evaluate_dict.items()
    }  # type: ignore # 保留4位小数


def glance_evaluate(
    frame: pl.LazyFrame | pl.DataFrame,
    start_date: str,
    end_date: str,
    created_date: str,
    columns_dict: dict[str, str] = {
        "equity": "equity",
        "time_return": "time_return",
        "drawdown": "drawdown",
    },
) -> GlanceEvaluateData:
    """概览评价回测结果"""
    res = frame.lazy() if isinstance(frame, pl.DataFrame) else frame
    equity_field = columns_dict.get("equity", "equity")
    time_return_field = columns_dict.get("time_return", "time_return")
    drawdown_field = columns_dict.get("drawdown", "drawdown")
    this_year_start_date: date = datetime.now().date().replace(month=1, day=1)
    this_month_start_date: date = datetime.now().date().replace(day=1)

    exprs = [
        pl.col(equity_field).last().alias("cum_asset"),  # 累计资产
        pl.col(equity_field).first().alias("first_equity"),  # 首日资产
        pl.col(time_return_field).first().alias("first_return"),  # 首日回报
        pl.len().alias("trading_days"),  # 交易天数
        pl.col(time_return_field)
        .mean()
        .replace(0, EPSILON)
        .alias("daily_pnl_mean"),  # 日盈亏均值
        pl.col(time_return_field)
        .std(ddof=1)
        .replace(0, EPSILON)
        .fill_null(EPSILON)
        .alias("daily_pnl_std"),  # 日盈亏标准差
        pl.col(drawdown_field)
        .min()
        .replace(0, EPSILON)
        .alias("daily_max_drawdown"),  # 日最大回撤
    ]
    # 计算基础统计量
    (
        stats,
        last_1_return_df,
        last_5_return_df,
        last_20_return_df,
        last_60_return_df,
        last_242_return_df,
        this_month_return_df,
        this_year_return_df,
        created_return_df,
    ) = pl.collect_all(
        [
            res.select(exprs),
            res.select(pl.col(time_return_field).get(-1)),
            res.select(
                pl.col(equity_field).get(-1) / pl.col(equity_field).shift(5).get(-1) - 1
            ),
            res.select(
                pl.col(equity_field).get(-1) / pl.col(equity_field).shift(20).get(-1)
                - 1
            ),
            res.select(
                pl.col(equity_field).get(-1) / pl.col(equity_field).shift(60).get(-1)
                - 1
            ),
            res.select(
                pl.col(equity_field).get(-1) / pl.col(equity_field).shift(242).get(-1)
                - 1
            ),
            res.select(
                (
                    pl.col(equity_field).last()
                    / pl.coalesce(
                        pl.col(equity_field)
                        .filter(pl.col("trade_date") < this_month_start_date)
                        .last(),
                        pl.col(equity_field).first(),
                    )
                )
                - 1
            ),
            res.select(
                (
                    pl.col(equity_field).last()
                    / pl.coalesce(
                        pl.col(equity_field)
                        .filter(pl.col("trade_date") < this_year_start_date)
                        .last(),
                        pl.col(equity_field).first(),
                    )
                )
                - 1
            ),
            res.select(
                pl.col(equity_field).get(-1)
                / pl.col(equity_field)
                .filter(
                    pl.col("trade_date")
                    >= datetime.strptime(created_date, "%Y-%m-%d").date()
                )
                .first()
                - 1
            ),
        ]
    )
    # 安全 round：值为 None 时保留 None（表示该指标无法计算），前端显示为 "-"
    def _round(v, n=4):
        return round(v, n) if v is not None else None

    # 计算衍生指标
    _first_equity = stats["first_equity"].item()
    _first_return = stats["first_return"].item()
    _cum_asset = stats["cum_asset"].item()
    _trading_days = stats["trading_days"].item()
    if _first_equity is None or _first_return is None or _cum_asset is None or not _trading_days:
        raise ValueError("回测结果为空，无法计算收益指标（策略可能在所有交易日均无有效选股）")
    capital: float = round(_first_equity / (_first_return + 1), 2)
    total_return: float = _cum_asset / capital - 1
    annu_return: float = (total_return + 1) ** (PERIODS / _trading_days) - 1
    _daily_max_drawdown = stats["daily_max_drawdown"].item()
    _pnl_mean = stats["daily_pnl_mean"].item()
    _pnl_std = stats["daily_pnl_std"].item()
    sharpe_ratio = (_pnl_mean / _pnl_std) * np.sqrt(PERIODS) if _pnl_mean is not None and _pnl_std else None

    evaluate_dict: GlanceEvaluateData = {
        "start_date": start_date,
        "end_date": end_date,
        "created_date": created_date,
        "annu_return": _round(annu_return),
        "daily_max_drawdown": _round(_daily_max_drawdown),
        "sharpe_ratio": _round(sharpe_ratio),
        "last_1_return": _round(last_1_return_df.item()),
        "last_5_return": _round(last_5_return_df.item()),
        "last_20_return": _round(last_20_return_df.item()),
        "last_60_return": _round(last_60_return_df.item()),
        "last_242_return": _round(last_242_return_df.item()),
        "this_month_return": _round(this_month_return_df.item()),
        "this_year_return": _round(this_year_return_df.item()),
        "created_return": _round(created_return_df.item()),
        "total_return": _round(total_return),
    }
    return evaluate_dict



"""
因子分析核心计算逻辑（stk/cb/etf 共用）

主要职责：
1. 识别调仓日并计算各期持仓内因子值
2. 计算持有期收益（close 价格法）
3. 计算每期 Rank IC（Spearman 相关）
4. 汇总统计指标、滚动 IC 均值、累计 IC
5. 计算因子相关矩阵
"""

import re
from typing import Any, Generator

import polars as pl

from ldnode import logger
from ldnode.calculator.handlers.models import TaskResponse
from ldnode.calculator.handlers.utils import Timer, response
from ldnode.calculator.operators.correlation import _normalize_corr_value
from ldnode.calculator.operators import factor_correlation_matrix


_HOLD_SECURITY_CODE_PATTERN = re.compile(r"\(([^()]+)\)$")


def _normalize_security_code(security: str) -> str:
    """兼容 `名称(代码)` 与纯 `代码` 两种持仓标识格式，统一提取为纯代码。"""
    matched = _HOLD_SECURITY_CODE_PATTERN.search(security)
    return matched.group(1) if matched else security


def _normalize_holdings(holdings: list[str] | None) -> list[str]:
    """批量规范化持仓标识，供因子分析与相关矩阵计算统一使用。"""
    if not holdings:
        return []
    return [_normalize_security_code(security) for security in holdings]


def _build_factor_analysis_fact_table(
    res_df: pl.DataFrame,
    screen_df: pl.DataFrame,
    factor_keys: list[str],
) -> tuple[list, pl.DataFrame]:
    """
    构造因子分析统一事实表，避免按调仓期重复切片与 join。

    返回:
        valid_dates: 可计算 IC 的调仓日列表（最后一期无 next date，已剔除）
        fact_df: 列结构
            rebalance_date / next_trade_date / code / close_ti / close_tnext /
            holding_return / factor_keys...
    """
    all_rebalance_dates = _identify_rebalance_dates(res_df)
    if len(all_rebalance_dates) < 2:
        return [], pl.DataFrame()

    valid_dates = all_rebalance_dates[:-1]
    next_dates = all_rebalance_dates[1:]
    rebalance_pair_df = pl.DataFrame(
        {
            "rebalance_date": valid_dates,
            "next_trade_date": next_dates,
        }
    )

    # 使用全排名截面（rank 不为 null）作为 IC 计算范围，
    # 替代原来仅使用持仓标的的方式，使 IC 反映因子对全候选池的预测力（学术标准）
    ranked_securities_df = (
        screen_df
        .filter(
            pl.col("trade_date").is_in(valid_dates),
            pl.col("rank").is_not_null(),
        )
        .select(
            pl.col("trade_date").alias("rebalance_date"),
            pl.col("code").cast(pl.String),
        )
        .unique()
    )

    if ranked_securities_df.is_empty():
        return valid_dates, pl.DataFrame()

    current_factor_df = screen_df.select(
        [
            pl.col("trade_date").alias("rebalance_date"),
            pl.col("code").cast(pl.String),  # screen_df.code 为 cat，需与 normalized_holdings_df.code(str) 对齐
            pl.col("close").alias("close_ti"),
            *factor_keys,
        ]
    )
    next_close_df = screen_df.select(
        [
            pl.col("trade_date").alias("next_trade_date"),
            pl.col("code").cast(pl.String),  # 同上
            pl.col("close").alias("close_tnext"),
        ]
    )

    fact_df = (
        rebalance_pair_df
        .join(ranked_securities_df, on="rebalance_date", how="inner")
        .join(current_factor_df, on=["rebalance_date", "code"], how="inner")
        .join(next_close_df, on=["next_trade_date", "code"], how="inner")
        .with_columns(
            ((pl.col("close_tnext") / pl.col("close_ti")) - 1.0).alias("holding_return")
        )
        .sort(["rebalance_date", "code"])
    )
    return valid_dates, fact_df


def _compute_factor_analysis(
    res_df: pl.DataFrame,
    screen_df: pl.DataFrame,
    factor_keys: list[str],
    rolling_period: int,
    task_id: str,
    log_prefix: str,
) -> Generator[TaskResponse[Any], None, None]:
    """
    因子分析核心计算，生成三帧输出。

    Args:
        res_df: 回测结果 DataFrame，含 trade_date / hold_securities 等列
        screen_df: 完整截面 DataFrame，含 code / trade_date / close / 各因子列
        factor_keys: 需要分析的因子名称列表
        rolling_period: 滚动 IC 均值窗口期数
        task_id: 任务 ID（用于日志）
        log_prefix: 日志前缀（如 "股票因子分析"）
    """
    with Timer(f"{log_prefix} 识别调仓日. 任务ID: {task_id}", logger=logger.event_backtest):
        rebalance_dates, ic_by_factor = _calc_ic_series(res_df, screen_df, factor_keys)

    with Timer(f"{log_prefix} 汇总统计指标. 任务ID: {task_id}", logger=logger.event_backtest):
        ic_stats = _calc_ic_stats(ic_by_factor, factor_keys)
        ic_series_data = _calc_ic_series_data(
            rebalance_dates, ic_by_factor, factor_keys, rolling_period
        )
        corr_matrix = _calc_factor_correlation(screen_df, res_df, factor_keys)

    # 帧1：各因子 IC 统计指标
    yield response(
        data_type="factor_ic_stats",
        data={"factors": ic_stats},
    )

    # 帧2：IC 时序数据（用于图表）
    yield response(
        data_type="factor_ic_series",
        data={
            "rebalance_dates": [d.strftime("%Y-%m-%d") for d in rebalance_dates],
            "factors": ic_series_data,
        },
    )

    # 帧3：因子相关矩阵（下三角）
    yield response(
        data_type="factor_correlation_matrix",
        data=corr_matrix,
    )


def _identify_rebalance_dates(res_df: pl.DataFrame) -> list:
    """
    识别调仓日：hold_securities 从 T-1 到 T 发生变化（或第一天）的日期列表。

    res_df 需含 trade_date（date 类型或字符串）、hold_securities（List[str]）列。
    返回已排序的日期对象列表。
    """
    # 按 trade_date 升序排列
    df = res_df.sort("trade_date")
    dates = df["trade_date"].to_list()
    holdings = df["hold_securities"].to_list()

    rebalance_dates = []
    prev_holdings = None
    for d, h in zip(dates, holdings):
        normalized_holdings = _normalize_holdings(h)
        h_set = set(normalized_holdings)
        prev_set = set(_normalize_holdings(prev_holdings))
        # 第一天直接算调仓日；或持仓集合发生变化时
        if prev_holdings is None or h_set != prev_set:
            rebalance_dates.append(d)
        prev_holdings = normalized_holdings

    return rebalance_dates


def _calc_ic_series(
    res_df: pl.DataFrame,
    screen_df: pl.DataFrame,
    factor_keys: list[str],
) -> tuple[list, dict[str, list[float | None]]]:
    """
    计算各调仓期的 Rank IC（Spearman 相关）。

    返回:
        rebalance_dates: 有效的调仓日列表（最后一个调仓期因无 T_next 而被剔除）
        ic_by_factor: {factor_key: [ic_T1, ic_T2, ...]}
    """
    valid_dates, fact_df = _build_factor_analysis_fact_table(res_df, screen_df, factor_keys)
    if not valid_dates:
        return [], {k: [] for k in factor_keys}

    ic_by_factor: dict[str, list[float | None]] = {k: [] for k in factor_keys}
    fact_by_date = {
        key[0]: frame
        for key, frame in fact_df.partition_by("rebalance_date", as_dict=True).items()
    }

    for t_i in valid_dates:
        merged = fact_by_date.get(t_i)
        if merged is None or len(merged) < 3:
            # 无持仓，IC 设为 None
            for k in factor_keys:
                ic_by_factor[k].append(None)
            continue

        holding_returns = merged["holding_return"].to_list()

        # 对每个因子计算 Spearman IC
        for k in factor_keys:
            if k not in merged.columns:
                ic_by_factor[k].append(None)
                continue
            factor_vals = merged[k].to_list()

            # 过滤 None / NaN
            pairs = [
                (f, r)
                for f, r in zip(factor_vals, holding_returns)
                if f is not None and r is not None
                and f == f and r == r  # NaN 自比较为 False
            ]
            if len(pairs) < 3:
                ic_by_factor[k].append(None)
                continue

            fv, rv = zip(*pairs)
            # Polars Spearman：方差为 0 时返回 null → Python None
            ic_df = pl.DataFrame({"f": list(fv), "r": list(rv)})
            ic_by_factor[k].append(
                _normalize_corr_value(
                    ic_df.select(pl.corr("f", "r", method="spearman")).item()
                )
            )

    return valid_dates, ic_by_factor


def _calc_ic_stats(
    ic_by_factor: dict[str, list[float | None]],
    factor_keys: list[str],
) -> list[dict]:
    """汇总每个因子的 IC 统计指标"""
    result = []
    for k in factor_keys:
        ic_list = ic_by_factor.get(k, [])
        # 只保留非 None 的有效 IC 值
        valid_ic = [v for v in ic_list if v is not None]

        if not valid_ic:
            result.append({
                "factor": k,
                "current_ic": None,
                "ic_mean": None,
                "ic_std": None,
                "ir": None,
                "ic_win_rate": None,
                "ic_max": None,
                "ic_min": None,
                "ic_skew": None,
                "ic_kurt": None,
            })
            continue

        ic_arr = pl.Series(valid_ic)
        ic_mean = float(ic_arr.mean())
        ic_std = float(ic_arr.std(ddof=1)) if len(valid_ic) > 1 else None
        ir = (ic_mean / ic_std) if ic_std else None
        _skew = ic_arr.skew()
        _kurt = ic_arr.kurtosis()

        result.append({
            "factor": k,
            # current_ic = 最后一个有效 IC（对应最后一个有数据的调仓期）
            "current_ic": valid_ic[-1],
            "ic_mean": ic_mean,
            "ic_std": ic_std,
            "ir": ir,
            "ic_win_rate": float(sum(1 for v in valid_ic if v > 0) / len(valid_ic)),
            "ic_max": float(ic_arr.max()),
            "ic_min": float(ic_arr.min()),
            "ic_skew": float(_skew) if _skew is not None else None,
            "ic_kurt": float(_kurt) if _kurt is not None else None,
        })

    return result


def _calc_ic_series_data(
    rebalance_dates: list,
    ic_by_factor: dict[str, list[float | None]],
    factor_keys: list[str],
    rolling_period: int,
) -> dict[str, dict]:
    """
    计算每个因子的 IC 时序数据：
    - ic: 原始 IC 序列
    - rolling_ic_mean: 滚动均值（前 rolling_period-1 期为 None）
    - cumulative_ic: 累计 IC（前缀和）
    """
    result = {}
    n = len(rebalance_dates)

    for k in factor_keys:
        ic_list = ic_by_factor.get(k, [None] * n)

        # 滚动 IC 均值
        rolling = []
        for i in range(n):
            window = ic_list[max(0, i - rolling_period + 1): i + 1]
            if len(window) < rolling_period:
                rolling.append(None)
            else:
                # None 视为 0 参与均值，分母固定为 rolling_period（而非 valid count）
                # 边界：rolling_period=1 时窗口始终满足，None IC 输出 0.0（非 None）
                total = sum(v if v is not None else 0.0 for v in window)
                rolling.append(total / rolling_period)

        # 累计 IC（前缀和，None 值贡献 0）
        cumulative = []
        running = 0.0
        for v in ic_list:
            running += v if v is not None else 0.0
            cumulative.append(running)

        result[k] = {
            "ic": ic_list,
            "rolling_ic_mean": rolling,
            "cumulative_ic": cumulative,
        }

    return result


def _calc_factor_correlation(
    screen_df: pl.DataFrame,
    res_df: pl.DataFrame,
    factor_keys: list[str],
) -> dict:
    """
    计算因子 Spearman 相关矩阵（基于全时段持仓内因子值）。

    方法：收集所有调仓日持仓内的因子值，rank 后计算 Pearson 相关矩阵。
    返回下三角矩阵（i > j 有值，i == j 为 1.0，i < j 为 None）。
    """
    n = len(factor_keys)

    if n == 0:
        return {"labels": [], "matrix": []}

    if n == 1:
        return {"labels": factor_keys, "matrix": [[1.0]]}

    _, fact_df = _build_factor_analysis_fact_table(res_df, screen_df, factor_keys)
    if fact_df.is_empty():
        # 无数据，返回对角矩阵
        matrix = _build_lower_triangle_matrix(
            [[1.0 if i == j else None for j in range(n)] for i in range(n)]
        )
        return {"labels": factor_keys, "matrix": matrix}

    return factor_correlation_matrix(fact_df.select(factor_keys), factor_keys)


def _build_lower_triangle_matrix(raw: list[list]) -> list[list]:
    """保持下三角矩阵格式（i < j 设为 None）"""
    n = len(raw)
    result = []
    for i in range(n):
        row = []
        for j in range(n):
            if j > i:
                row.append(None)
            else:
                row.append(raw[i][j])
        result.append(row)
    return result



import asyncio as aio
import gzip
import json
import multiprocessing as mp
import queue

import aiohttp
import bson
import humanize

from ldnode import logger
from ldnode.loader.persistence.light_cos import LightCOSClient
from ldnode.runner.config import Settings
from ldnode.runner.events import TaskAnswer
from ldnode.runner.schemas import Worker
from ldnode.runner.task_manager import TaskManager

from .utils import codec_options


class TaskAnswerCompletionLockState:
    """任务应答完成锁"""

    def __init__(self):
        self.waiting_answers: dict[str, int] = {}
        self.completed_answers: dict[str, tuple[TaskAnswer, aiohttp.FormData, int]] = {}
        self.locks: dict[str, aio.Lock] = {}

    async def pre_register(self, task_id: str) -> None:
        """非 completed 帧在 COS 上传前预先占位，防止 completed 帧在 COS 上传期间抢先发送"""
        if task_id not in self.locks:
            self.locks[task_id] = aio.Lock()
        async with self.locks[task_id]:
            self.waiting_answers[task_id] = self.waiting_answers.get(task_id, 0) + 1

    async def is_completed(
        self, answer: TaskAnswer, form_data: aiohttp.FormData, payload_size: int
    ) -> bool:
        # pre_register 已在 COS 上传前完成计数，此处非 completed 帧直接发送
        if answer["task_id"] not in self.locks:
            self.locks[answer["task_id"]] = aio.Lock()
        async with self.locks[answer["task_id"]]:
            if answer["completed"]:
                if self.waiting_answers.get(answer["task_id"], 0) <= 0:
                    return True
                else:
                    self.completed_answers[answer["task_id"]] = (
                        answer,
                        form_data,
                        payload_size,
                    )
                    return False
            else:
                # waiting_answers 已由 pre_register 计数，无需再次递增
                return True

    async def complete(
        self, answer: TaskAnswer
    ) -> tuple[TaskAnswer, aiohttp.FormData, int] | None:
        if answer["task_id"] not in self.locks:
            self.locks[answer["task_id"]] = aio.Lock()
        async with self.locks[answer["task_id"]]:
            if not answer["completed"]:
                if answer["task_id"] in self.waiting_answers:
                    self.waiting_answers[answer["task_id"]] -= 1
                if self.waiting_answers[answer["task_id"]] <= 0:
                    del self.waiting_answers[answer["task_id"]]
                    if answer["task_id"] in self.completed_answers:
                        return self.completed_answers.pop(answer["task_id"])
            del self.locks[answer["task_id"]]


class PostWorker(Worker):
    """后端工作进程"""

    ANSWER_URI: str = "/v1/calculation/node-answer"

    def __init__(self, worker_name: str):
        super().__init__(worker_name)
        self.settings: Settings | None = None
        self.inq: mp.Queue | None = None
        self.outq: mp.Queue | None = None
        self.tasks: TaskManager = TaskManager()
        self.http_session: aiohttp.ClientSession | None = None
        self.cos_client: LightCOSClient | None = None
        self.task_answer_completion_lock_state = TaskAnswerCompletionLockState()

    async def post_task_answer(
        self,
        data: TaskAnswer,
        form_data: aiohttp.FormData,
        payload_size: int,
        retry_count: int = 0,
    ) -> None:
        """发送任务应答"""
        if not self.http_session:
            return
        is_retry = False
        try:
            if retry_count > 3:
                logger.connector.error(
                    f"上报({data['method']})任务({data['task_id']})数据失败: 重试次数超过3次"
                )
                return
            response = await self.http_session.post(self.ANSWER_URI, data=form_data)
            match response.status:
                case 200:
                    logger.connector.info(
                        f"上报({data['method']})任务({data['task_id']})数据成功 大小: {humanize.naturalsize(payload_size)}"
                    )
                    if (
                        completed_answer
                        := await self.task_answer_completion_lock_state.complete(data)
                    ):
                        _data, _form_data, _payload_size = completed_answer
                        await self.post_task_answer(_data, _form_data, _payload_size)
                    return
                case 401:
                    logger.connector.error(
                        f"上报({data['method']})任务({data['task_id']})数据失败: 认证失败"
                    )
                    return
                case _:
                    logger.connector.warning(
                        f"上报({data['method']})任务({data['task_id']})数据失败: HTTP Error {response.status}, 重试 {retry_count + 1} / 3"
                    )
                    is_retry = True
        except aiohttp.ConnectionTimeoutError:
            logger.connector.warning(
                f"上报({data['method']})任务({data['task_id']})数据失败: 连接超时, 重试 {retry_count + 1} / 3"
            )
            is_retry = True
        except aiohttp.ServerConnectionError:
            logger.connector.warning(
                f"上报({data['method']})任务({data['task_id']})数据失败: 服务端连接失败, 疑似服务端不在线, 重试 {retry_count + 1} / 3"
            )
            is_retry = True
        except Exception as e:
            logger.connector.warning(
                f"上报({data['method']})任务({data['task_id']})数据失败: {e}, 重试 {retry_count + 1} / 3"
            )
            is_retry = True
        finally:
            if is_retry:
                await aio.sleep(retry_count)
                await self.tasks.create_task(
                    self.post_task_answer(
                        data, form_data, payload_size, retry_count + 1
                    )
                )

    async def post_to_cos(self, key: str, content: bytes | dict) -> None:
        """上传文件到COS"""
        if self.cos_client:
            if type(content) is bytes:
                bt_content = content
            else:
                bt_content = json.dumps(content).encode()
            await aio.to_thread(
                self.cos_client.upload_file, key=key, content=bt_content
            )

    async def post_answer(self, data: TaskAnswer) -> None:
        """发送任务应答"""
        if not self.settings or not self.cos_client:
            return
        # 非 completed 帧先占位：防止 completed 帧在 COS 上传期间因计数归零而抢先发送
        if not data["completed"]:
            await self.task_answer_completion_lock_state.pre_register(data["task_id"])
        if storage_key := data["result"].get("storage_key"):
            # 上传数据到COS
            await self.post_to_cos(storage_key, data["result"]["data"])
            # 并设置返回格式
            data_result = {
                "data_type": data["result"]["data_type"],
                "data": {
                    "storage_key": storage_key,
                },
            }
        else:
            data_result = data["result"]

        result: bytes = bson.encode(data_result, codec_options=codec_options)
        result = gzip.compress(result, compresslevel=6)
        form_data = aiohttp.FormData()
        form_data.add_field("result", result, content_type="application/gzip")
        form_data.add_field("task_id", data["task_id"])
        form_data.add_field("completed", str(data["completed"]))
        form_data.add_field("method", data["method"])
        form_data.add_field("access_id", self.settings.ENDPOINT_ACCESS_ID)
        form_data.add_field("access_token", self.settings.ENDPOINT_ACCESS_TOKEN)
        if await self.task_answer_completion_lock_state.is_completed(
            data, form_data, len(result)
        ):
            await self.post_task_answer(data, form_data, len(result))

    async def _run(self) -> None:
        """运行工作进程"""
        if not self.settings or not self.inq:
            return
        self.http_session = aiohttp.ClientSession(
            base_url=self.settings.ENDPOINT_URL.replace("ws://", "http://").replace(
                "wss://", "https://"
            )
        )
        self.cos_client = LightCOSClient(
            region=self.settings.RESULT_COS_REGION,
            secret_id=self.settings.RESULT_COS_SECRET_ID,
            secret_key=self.settings.RESULT_COS_SECRET_KEY,
            bucket=self.settings.RESULT_COS_BUCKET,
        )
        while self.running:
            try:
                data = await aio.to_thread(self.inq.get, timeout=1)
                if data is None:
                    self.running = False
                    break
                self.tasks.create_task(self.post_answer(data))
            except queue.Empty:
                continue
            except Exception as e:
                logger.connector.error(f"监听输入队列失败: {e}")
                continue
        await self.http_session.close()

    def run(self, settings: Settings, input: mp.Queue, output: mp.Queue):
        """运行工作进程"""
        self.settings = settings
        self.inq = input
        self.outq = output
        aio.run(self._run())



import polars as pl
from polars import col


def plan_correct(lf: pl.LazyFrame) -> pl.LazyFrame:
    """ 数据纠正计划 """
    return (
        lf.with_columns(
            col('code').str.slice(2) + '.' + col('code').str.slice(0, 2).str.to_uppercase(),
            col('trade_date').str.to_date(),
            col('time').str.to_time(),
            pl.when((col('bid1')==0) | (col('ask1')==0)) # 修正涨跌停
            .then(col('bid1') + col('ask1'))
            .otherwise((col('bid1') + col('ask1')) / 2)
            .alias('close'),
        )
        .with_columns(
            # 修正无挂单close
            pl.when(col('close')==0)
            .then(
                pl.when(col('now')==0).then(col('pre_close'))# 开盘无挂单修正
                .otherwise(col('now'))
            ).otherwise(col('close')).alias('close'),
        )
    )


from datetime import date
import math

import polars as pl
import pytest

from ldnode.calculator.handlers._factor_analysis_core import (
    _build_factor_analysis_fact_table,
    _calc_factor_correlation,
    _calc_ic_series,
    _calc_ic_series_data,
    _calc_ic_stats,
    _identify_rebalance_dates,
    _normalize_holdings,
    _normalize_security_code,
)


def _build_res_df() -> pl.DataFrame:
    return pl.DataFrame(
        {
            "trade_date": [date(2024, 1, 1), date(2024, 1, 2), date(2024, 1, 3)],
            "hold_securities": [
                ["Alpha(AAA)", "Beta(BBB)", "Gamma(CCC)"],
                ["Alpha(AAA)", "Gamma(CCC)", "Delta(DDD)"],
                ["Alpha(AAA)", "Gamma(CCC)", "Delta(DDD)"],
            ],
        }
    )


def _build_screen_df() -> pl.DataFrame:
    return pl.DataFrame(
        {
            "trade_date": [
                date(2024, 1, 1),
                date(2024, 1, 1),
                date(2024, 1, 1),
                date(2024, 1, 2),
                date(2024, 1, 2),
                date(2024, 1, 2),
                date(2024, 1, 3),
                date(2024, 1, 3),
                date(2024, 1, 3),
            ],
            "code": ["AAA", "BBB", "CCC", "AAA", "BBB", "CCC", "AAA", "CCC", "DDD"],
            "close": [10.0, 10.0, 10.0, 11.0, 12.0, 13.0, 12.0, 14.0, 16.0],
            "factor_a": [1.0, 2.0, 3.0, 5.0, 4.0, 6.0, 7.0, 8.0, 9.0],
            "factor_b": [3.0, 2.0, 1.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
            "factor_constant": [1.0] * 9,
        }
    )


def test_identify_rebalance_dates_only_marks_first_day_and_holdings_change() -> None:
    res_df = _build_res_df()

    result = _identify_rebalance_dates(res_df)

    assert result == [date(2024, 1, 1), date(2024, 1, 2)]


def test_normalize_security_code_supports_plain_code_and_name_code() -> None:
    assert _normalize_security_code("123013.SZ") == "123013.SZ"
    assert _normalize_security_code("横河转债(123013.SZ)") == "123013.SZ"
    assert _normalize_holdings(["横河转债(123013.SZ)", "113576.SH"]) == [
        "123013.SZ",
        "113576.SH",
    ]


def test_calc_ic_series_returns_expected_rank_ic_values() -> None:
    res_df = _build_res_df()
    screen_df = _build_screen_df()

    rebalance_dates, ic_by_factor = _calc_ic_series(
        res_df,
        screen_df,
        ["factor_a", "factor_b", "factor_constant"],
    )

    assert rebalance_dates == [date(2024, 1, 1)]
    assert ic_by_factor["factor_a"] == pytest.approx([1.0])
    assert ic_by_factor["factor_b"] == pytest.approx([-1.0])
    assert ic_by_factor["factor_constant"] == [None]


def test_build_factor_analysis_fact_table_reuses_normalized_holdings_and_computes_returns() -> None:
    valid_dates, fact_df = _build_factor_analysis_fact_table(
        _build_res_df(),
        _build_screen_df(),
        ["factor_a", "factor_b"],
    )

    assert valid_dates == [date(2024, 1, 1)]
    assert fact_df.select("code").to_series().to_list() == ["AAA", "BBB", "CCC"]
    assert fact_df.select("holding_return").to_series().to_list() == pytest.approx(
        [0.1, 0.2, 0.3]
    )
    assert fact_df.columns == [
        "rebalance_date",
        "next_trade_date",
        "code",
        "close_ti",
        "factor_a",
        "factor_b",
        "close_tnext",
        "holding_return",
    ]


def test_calc_ic_stats_uses_last_valid_ic_and_ignores_none() -> None:
    result = _calc_ic_stats(
        {
            "factor_a": [None, 0.5, -0.5],
            "factor_b": [None, None],
        },
        ["factor_a", "factor_b"],
    )

    factor_a = next(item for item in result if item["factor"] == "factor_a")
    factor_b = next(item for item in result if item["factor"] == "factor_b")

    assert factor_a["current_ic"] == -0.5
    assert factor_a["ic_mean"] == 0.0
    assert math.isclose(factor_a["ic_std"], math.sqrt(0.5), rel_tol=1e-9)
    assert factor_a["ir"] == 0.0
    assert factor_a["ic_win_rate"] == 0.5
    assert factor_a["ic_max"] == 0.5
    assert factor_a["ic_min"] == -0.5
    assert factor_b == {
        "factor": "factor_b",
        "current_ic": None,
        "ic_mean": None,
        "ic_std": None,
        "ir": None,
        "ic_win_rate": None,
        "ic_max": None,
        "ic_min": None,
        "ic_skew": None,
        "ic_kurt": None,
    }


def test_calc_ic_series_data_builds_rolling_mean_and_cumulative_sum() -> None:
    result = _calc_ic_series_data(
        [date(2024, 1, 1), date(2024, 1, 2), date(2024, 1, 3)],
        {"factor_a": [1.0, None, -0.5]},
        ["factor_a"],
        rolling_period=2,
    )

    assert result == {
        "factor_a": {
            "ic": [1.0, None, -0.5],
            "rolling_ic_mean": [None, 1.0, -0.5],
            "cumulative_ic": [1.0, 1.0, 0.5],
        }
    }


def test_calc_factor_correlation_collects_rebalance_holdings_and_returns_lower_triangle() -> None:
    result = _calc_factor_correlation(
        _build_screen_df(),
        _build_res_df(),
        ["factor_a", "factor_b", "factor_constant"],
    )

    assert result["labels"] == ["factor_a", "factor_b", "factor_constant"]
    assert result["matrix"][0] == [1.0, None, None]
    assert result["matrix"][1][0] < 0
    assert result["matrix"][1][1] == 1.0
    assert result["matrix"][2] == [None, None, 1.0]



from ldnode.calculator.schemas import ComputedFactor
from .common import TaskBaseParameters


class FormulaComputeParameters(TaskBaseParameters):
    """ 可转债公式计算参数 """
    formula_list: list[list[ComputedFactor]]



import bisect
from functools import reduce

import pandas as pd
import numpy as np

from ldnode.calculator.schemas import GridBacktestParameters, GridType


def cb_grid_backtest(
    df: pd.DataFrame,
    strategy_config: GridBacktestParameters,
) -> pd.DataFrame:
    """
    网格回测
    """
    df.set_index('datetime', inplace=True)
    df.dropna(inplace=True)
    # 初始化网格
    if strategy_config['grid_type'] == GridType.LINE:
        grid = list(np.linspace(strategy_config['grid_range'][0], strategy_config['grid_range'][1], strategy_config['grid_count'] + 1).round(3))
    elif strategy_config['grid_type'] == GridType.GEOM:
        grid = list(np.geomspace(strategy_config['grid_range'][0], strategy_config['grid_range'][1], strategy_config['grid_count'] + 1).round(3))
    else:
        raise ValueError(f"网格类型错误: {strategy_config['grid_type']}")
    # 生成所需
    df['preclose'] = df.close.shift(1)
    df['inc'] = df.close - df.preclose
    df['grid_index'] = None
    df['pos_chg'] = None
    df['pos'] = None
    df['pos_val'] = None
    df['cash'] = None
    df['grid_return'] = None
    df['grid_return_acc'] = None
    df['asset'] = None
    # 定位网格初始位置 投入底仓
    grid_index = df.columns.get_loc('grid_index')
    inc = df.columns.get_loc('inc')
    open_price = df.columns.get_loc('open')
    close = df.columns.get_loc('close')
    pos = df.columns.get_loc('pos')
    cash = df.columns.get_loc('cash')
    grid_return = df.columns.get_loc('grid_return')
    df.iat[0, grid_index] = grid.index(min(grid, key=lambda x: abs(x - df.iat[0, 0]))) # 当前位置
    df.iat[0, pos] = ((len(grid) - 1) - df.iat[0, grid_index]) * 10 # 投入底仓
    df.iat[0, cash] = -(df.iat[0, pos] * df.iat[0, open_price])# 借入现金

    # 计算网格位置
    for i in range(1, df.shape[0]):
        top_flag = False # 触碰网格顶部
        bottom_flag = False # 触碰网格底部
        pre_index = df.iat[i - 1, grid_index] # 上周期网格所处位置
        if pre_index == (len(grid) - 1):
            # print('向上破网\n', df.iloc[[i]])
            if strategy_config['stop_on_top']:
                break
            top_flag = True
        elif pre_index == 0:
            # print('向下破网，不再买入\n', df.iloc[[i]].to_string())
            bottom_flag = True
        pre_cash = df.iat[i - 1, cash] # 上周期现金
        upper_con = (df.iat[i, inc] > 0) & (df.iat[i, close] >= grid[pre_index + 1]) if not top_flag else False
        lower_con = (df.iat[i, inc] < 0) & (df.iat[i, close] <= grid[pre_index - 1]) if not bottom_flag else False
        # 网格上移
        if upper_con:
            index = (bisect.bisect_right(grid, df.iat[i, close]) - 1) # 上移后index位置
            df.iat[i, grid_index] = index # 写入当前index位置
            l = grid[pre_index + 1 : index + 1] # 上移经过位置
            cost = reduce(lambda x, y: x + y, l)
            df.iat[i, cash] = pre_cash + 10 * cost * (1 - strategy_config['grid_c_rate']) # 卖出变现
            # 纯网格收益
            df.iat[i, grid_return] = sum(map(lambda x : (grid[grid.index(x)] - grid[grid.index(x) - 1]) , l)) * 10 * (1 - strategy_config['grid_c_rate']) ** 2
        # 网格下移
        elif lower_con:
            index = bisect.bisect_left(grid ,df.iat[i, close]) # 下移后index位置
            df.iat[i, grid_index] = index # 写入当前index位置
            l = grid[index:pre_index] # 下移经过位置
            cost = reduce(lambda x, y: (x + y), l)
            df.iat[i, cash] = pre_cash - 10 * cost * (1 + strategy_config['grid_c_rate'])# 借入现金
        # 网格不变
        else:
            df.iat[i, grid_index] = pre_index
            df.iat[i, cash] = pre_cash # 现金保持不变

    # 生成资产数据
    df.dropna(subset=['cash'], inplace=True) # 删除停止之后的数据
    df['pos_chg'] = (df.grid_index.shift(1) - df.grid_index) * 10 # 持仓变化
    df.loc[:,'pos'] = df.iat[0, pos] + df.pos_chg.expanding().sum() # 累计持仓
    df['pos_val'] = df.close * df.pos
    df.grid_return = df.grid_return.fillna(0)
    df['grid_return_acc'] = df.grid_return.expanding().sum()
    df['asset'] = df.pos_val + df.cash
    return df



from datetime import date
from types import SimpleNamespace

import polars as pl
import pytest

from ldnode.calculator.handlers.stk_factor_analysis_handler import (
    handle_stk_factor_analysis,
)
from ldnode.calculator.handlers.strategy_correlation_handler import (
    handle_strategy_correlation_analysis,
)
from ldnode.calculator.schemas.enumerations import SecurityType


def _build_strategy_res_df(time_returns: list[float], holdings: list[list[str]]) -> pl.DataFrame:
    return pl.DataFrame(
        {
            "trade_date": [date(2024, 1, d) for d in range(1, len(time_returns) + 1)],
            "time_return": time_returns,
            "hold_securities": holdings,
        }
    )


def _build_factor_res_df() -> pl.DataFrame:
    return pl.DataFrame(
        {
            "trade_date": [date(2024, 1, 1), date(2024, 1, 2), date(2024, 1, 3)],
            "hold_securities": [
                ["AAA", "BBB", "CCC"],
                ["AAA", "CCC", "DDD"],
                ["AAA", "CCC", "DDD"],
            ],
        }
    )


def _build_factor_screen_df() -> pl.DataFrame:
    return pl.DataFrame(
        {
            "trade_date": [
                date(2024, 1, 1),
                date(2024, 1, 1),
                date(2024, 1, 1),
                date(2024, 1, 2),
                date(2024, 1, 2),
                date(2024, 1, 2),
                date(2024, 1, 3),
                date(2024, 1, 3),
                date(2024, 1, 3),
            ],
            "code": ["AAA", "BBB", "CCC", "AAA", "CCC", "DDD", "AAA", "CCC", "DDD"],
            "close": [10.0, 11.0, 12.0, 13.0, 15.0, 17.0, 14.0, 16.0, 18.0],
            "factor_a": [1.0, 2.0, 3.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0],
            "factor_b": [3.0, 2.0, 1.0, 1.0, 0.0, -1.0, -2.0, -3.0, -4.0],
        }
    )


def test_strategy_correlation_handler_yields_three_business_frames_and_fee(monkeypatch) -> None:
    import ldnode.calculator.handlers.strategy_correlation_handler as strategy_handler

    res_df_map = {
        "alpha": _build_strategy_res_df(
            [0.01, 0.02, 0.03],
            [["000001", "000002"], ["000001", "000003"], ["000001", "000003"]],
        ),
        "beta": _build_strategy_res_df(
            [0.02, 0.04, 0.06],
            [["000002", "000004"], ["000001", "000003"], ["000001", "000003"]],
        ),
    }

    monkeypatch.setattr(
        strategy_handler,
        "_run_backtest",
        lambda entry, merged_config: res_df_map[entry["label"]],
    )

    results = list(
        handle_strategy_correlation_analysis(
            {
                "task_id": "task-strategy-handler",
                "global_settings": {
                    "start_date": "2024-01-01",
                    "end_date": "2024-01-31",
                    "initial_capital": 1_000_000.0,
                    "commission": 0.001,
                    "benchmark": "000300.SH",
                },
                "strategies": [
                    {
                        "security_type": SecurityType.STK,
                        "label": "alpha",
                        "core_config": {},
                    },
                    {
                        "security_type": SecurityType.STK,
                        "label": "beta",
                        "core_config": {},
                    },
                ],
                "rolling_period": 2,
            }
        )
    )

    assert [item["data_type"] for item in results] == [
        "correlation_series",
        "correlation_matrix",
        "overlap_matrix",
        "fee",
    ]
    assert results[0]["data"]["pairs"]["alpha+beta"][0] is None
    assert results[0]["data"]["pairs"]["alpha+beta"][1:] == pytest.approx([1.0, 1.0])
    assert results[1]["data"]["matrix"] == [[1.0, None], [1.0, 1.0]]
    assert results[2]["data"]["matrix"] == [[1.0, None], [7.0 / 9.0, 1.0]]


def test_stk_factor_analysis_handler_returns_error_when_rank_factors_empty() -> None:
    results = list(
        handle_stk_factor_analysis(
            {
                "task_id": "task-empty-rank-factors",
                "strategy_config": {"rank_factors": []},
                "rolling_period": 3,
            }
        )
    )

    assert [item["data_type"] for item in results] == ["error", "fee"]
    assert results[0]["data"] == "rank_factors 为空，无法进行因子分析"


def test_stk_factor_analysis_handler_yields_factor_frames_and_fee(monkeypatch) -> None:
    import ldnode.calculator.handlers.stk_factor_analysis_handler as stk_handler

    monkeypatch.setattr(
        stk_handler,
        "STK_DATA",
        SimpleNamespace(output_lazyframe=lambda: None),
    )
    monkeypatch.setattr(
        stk_handler,
        "stk_event_backtest",
        lambda **kwargs: (_build_factor_res_df(), None, _build_factor_screen_df()),
    )

    results = list(
        handle_stk_factor_analysis(
            {
                "task_id": "task-stk-factor",
                "strategy_config": {
                    "rank_factors": [
                        {"factor": "factor_a"},
                        {"factor": "factor_b"},
                    ]
                },
                "rolling_period": 2,
            }
        )
    )

    assert [item["data_type"] for item in results] == [
        "factor_ic_stats",
        "factor_ic_series",
        "factor_correlation_matrix",
        "fee",
    ]
    assert [item["factor"] for item in results[0]["data"]["factors"]] == [
        "factor_a",
        "factor_b",
    ]
    assert results[1]["data"]["rebalance_dates"] == ["2024-01-01"]
    assert results[2]["data"]["labels"] == ["factor_a", "factor_b"]



from datetime import date, time

import easyquotation
import polars as pl
import patito as pt

from ldnode import logger
from ldnode.loader.schemas.models import DataSource, SyncResult
from ldnode.loader.schemas.enumerations import SyncStatus
from ldnode.loader.files import ETF_PRE_QUOTATION, ETF_QUOTATION
from ldnode.loader.datasources.sina import SinaWithTimeoutPatch


class Model(pt.Model):
    trade_date: date # 交易日期
    time: time # 交易时间
    code: str # 标的代码
    name: str # 标的名称
    pre_close: float # 前收盘价(元)
    open: float # 开盘价(元)
    high: float # 最高价(元)
    low: float # 最低价(元)
    close: float # 收盘价(元)
    vol: int # 成交量(份)
    amount: float # 成交额(元)
    suspension: bool # 停牌标记

    cash_div: float = pt.Field(dtype=pl.Float32) # 现金分红(元/份)
    stk_div: float = pt.Field(dtype=pl.Float32) # 送股分红
    adj_factor: float # 复权因子
    adj_back: float # 后复权因子
    list_trade_days: int = pt.Field(dtype=pl.Int16) # 上市以来交易日天数
    
    list_days: int = pt.Field(dtype=pl.Int16) # 上市以来自然日天数
    pct_chg: float # 涨跌幅
    open_pct_chg: float # 开盘涨跌幅
    high_pct_chg: float # 最高涨跌幅
    low_pct_chg: float # 最低涨跌幅
    limit: int = pt.Field(dtype=pl.Int8) # 涨跌停标记
    
    last_trade_day: bool # 是否为最后一个交易日


class ETFQuotationsDataSource(DataSource):
    def load(self, key: str) -> list[SyncResult]:
        """ 加载 ETF 实时截面数据 """
        try:
            if not ETF_PRE_QUOTATION.get_local_path().exists():
                return [SyncResult(key, SyncStatus.FAILED, "数据源不存在")]
            quotation = SinaWithTimeoutPatch()  # 使用新浪数据源
            # 读取股票盘前数据
            pre_quotations = pl.scan_parquet(ETF_PRE_QUOTATION.get_local_path())
            # 下载行情数据
            quotation_data = quotation.stocks(
                [*map(
                    lambda code: code.split('.')[1].lower() + code.split('.')[0], 
                    pre_quotations.select(pl.col('code')).collect().to_series(),
                )],
                prefix=True
            )
            quotations = (
                pl.DataFrame(quotation_data)
                .transpose(include_header=True, header_name='code').lazy()
                .unnest('column_0')
                .rename({
                    'close': 'pre_close',
                    'turnover': 'vol',
                    'volume': 'amount',
                    'date': 'trade_date'
                })
                .with_columns(
                    pl.col('code').str.slice(2) + '.' + pl.col('code').str.slice(0, 2).str.to_uppercase(),
                    pl.col('trade_date').str.to_date(),
                    pl.col('time').str.to_time(),
                    pl.when((pl.col('bid1')==0) | (pl.col('ask1')==0)) # 修正涨跌停
                    .then(pl.col('bid1') + pl.col('ask1'))
                    .otherwise((pl.col('bid1') + pl.col('ask1')) / 2)
                    .alias('close'),
                )
                .with_columns(
                    # 修正无挂单close
                    pl.when(pl.col('close')==0)
                    .then(
                        pl.when(pl.col('now')==0).then(pl.col('pre_close'))# 开盘无挂单修正
                        .otherwise(pl.col('now'))
                    ).otherwise(pl.col('close')).alias('close'),
                )
                .join(
                    pre_quotations.rename({'close': '_pre_close'}), 
                    on='code', how='right'
                )
                .with_columns(
                    # 涨跌幅指标
                    (pl.col('close') / pl.col('pre_close') - 1).alias('pct_chg'),  # 涨跌幅
                    (pl.col("open") / pl.col("pre_close") - 1).alias("open_pct_chg"),  # 开盘涨跌幅
                    (pl.col("high") / pl.col("pre_close") - 1).alias("high_pct_chg"),  # 最高涨跌幅
                    (pl.col("low") / pl.col("pre_close") - 1).alias("low_pct_chg"),  # 最低涨跌幅
                    # 其他指标
                    (pl.col('_pre_close') - pl.col('pre_close')).fill_null(0).cast(pl.Float32).alias('cash_div'), # 现金分红(元/份)
                    pl.lit(0).cast(pl.Float32).alias('stk_div'), # 送股分红(元/份)
                    (pl.col('list_trade_days') + 1).fill_null(1).alias("list_trade_days"), # 上市交易日天数
                    (pl.col("trade_date") - pl.col("list_date")).dt.total_days().cast(pl.Int16).alias("list_days"),  # 上市自然日天数
                    (pl.col('adj_factor') * pl.col('close') / pl.col('pre_close')).fill_null(1).alias('adj_factor'), # 复权因子
                    (pl.col('_pre_close') / pl.col('pre_close') * pl.col('adj_back')).fill_null(1).alias('adj_back'), # 后复权因子
                    # 停牌标记
                    pl.when(
                        (pl.col("open") == 0) & (pl.col("bid1") == 0) & (pl.col("ask1") == 0)
                    ).then(True).otherwise(False).alias('suspension'),
                    pl.lit(0).cast(pl.Int8).alias('limit'), # 涨跌停标记
                    pl.lit(False).alias('last_trade_day'), # 是否为最后一个交易日
                )
                .select(Model.columns) # 保留所需列
                .collect() # 生成dataframe
                .pipe(Model.validate) # 数据验证
            )
            # 保存ETF实时截面数据
            quotations.write_parquet(ETF_QUOTATION.get_local_path())
            return [SyncResult(key, SyncStatus.SUCCESS, date.today().strftime("%Y-%m-%d"))]
        except Exception as e:
            logger.loader.error(f"加载 ETF 实时截面数据失败: {e}")
            return [SyncResult(key, SyncStatus.FAILED, str(e))]



from textwrap import shorten
from typing import Any, Optional

from ldnode.calculator.handlers.models import T, TaskResponse


def response(
    data_type: str, data: T, storage_key: Optional[str] = None
) -> TaskResponse[T]:
    """正常响应"""
    answer: TaskResponse[T] = {
        "data_type": data_type,
        "data": data,
    }
    if storage_key:
        answer["storage_key"] = storage_key
    return answer


def error_response(error_msg: Any) -> TaskResponse[str]:
    """错误响应"""
    return {
        "data_type": "error",
        "data": shorten(f"{error_msg}", width=255, placeholder="..."),
    }



from datetime import date
from typing import Optional

import polars as pl
import polars.selectors as cs

from ldnode.calculator.plans.polars import (
    planning_mod_rank,
    planning_screen_lf,
    planning_screen_weight,
)
from ldnode.calculator.schemas import (
    CB_BUILTIN_FACTORS,
    CB_SCREEN_RESULT_FIXED_COLUMNS,
    ETF_BUILTIN_FACTORS,
    ETF_SCREEN_RESULT_FIXED_COLUMNS,
    SCREEN_CB_SETTING_RELATIVE_FACTORS,
    SCREEN_COMMON_SETTING_RELATIVE_FACTORS,
    SCREEN_ETF_SETTING_RELATIVE_FACTORS,
    SCREEN_STK_SETTING_RELATIVE_FACTORS,
    STK_BUILTIN_FACTORS,
    STK_SCREEN_RESULT_FIXED_COLUMNS,
    BuiltinScope,
    CBBuiltinScope,
    CBNeutralizationMethod,
    CBProcessMissingDataMethod,
    CBScreenParameters,
    ETFBuiltinScope,
    ETFNeutralizationMethod,
    ETFProcessMissingDataMethod,
    ETFScreenParameters,
    ExcludeConditionGroup,
    NeutralizationMethod,
    ProcessMissingDataMethod,
    ScreenParameters,
    STKBuiltinScope,
    STKNeutralizationMethod,
    STKProcessMissingDataMethod,
    STKScreenParameters,
    gen_cb_filter_exprs,
    gen_etf_filter_exprs,
    gen_stk_filter_exprs,
)


def _gen_screen_result_columns(
    fixed_columns: list[str],
    parameters: ScreenParameters,
    setting_relative_factors: dict[str, list[str]],
    show_all_data: bool = False,
) -> list[str]:
    """生成返回数据列"""
    if show_all_data:
        # 显示所有列，base_columns 在前
        return (fixed_columns, cs.exclude(fixed_columns))
    # 计算要展示的自定义因子列
    customized_factors = []
    for fs in parameters.get("computed_factors", []):
        customized_factors.extend([f["factor"] for f in fs if not f["is_intermediate"]])
    # 计算快捷排除设置关联的因子列（合并通用映射 + 品种专属映射）
    merged_setting_factors = {}
    merged_setting_factors.update(SCREEN_COMMON_SETTING_RELATIVE_FACTORS)
    merged_setting_factors.update(setting_relative_factors)
    exclude_setting_factors: list[str] = []
    for setting, factors in merged_setting_factors.items():
        if parameters.get(setting):
            exclude_setting_factors.extend(factors)
    # 计算要展示的排除因子列
    if filter_factors := parameters.get("filter_factors"):
        exclude_factors = [
            f["factor"] for f in ExcludeConditionGroup.flatten(filter_factors)
        ]
    else:
        exclude_factors = []
    # 计算要展示的打分因子列
    rank_factors = [f["factor"] for f in parameters["rank_factors"]]
    # 计算要展示的打分列（命名规则与 score_factors.py 保持一致：有中性化时列名带 _neutralization）
    score_columns = [
        f"_{f['factor']}_neutralization_score" if f.get("neutralization_factor") else f"_{f['factor']}_score"
        for f in parameters["rank_factors"]
    ] + ["score"]
    # 返回列定义
    return cs.by_name(
        list(
            dict.fromkeys(
                fixed_columns
                + customized_factors
                + exclude_setting_factors
                + exclude_factors
                + rank_factors
                + score_columns
            )
        )
    )


def _planning_screen(
    screen_lf: pl.LazyFrame,
    trade_date: date,
    parameters: ScreenParameters,
    builtin_factors: list[str],
    BuiltinScopes: type[BuiltinScope],
    ProcessMissingDataMethods: type[ProcessMissingDataMethod],
    NeutralizationMethods: type[NeutralizationMethod],
    filter_exprs: list[pl.Expr],
    fixed_columns: list[str],
    setting_relative_factors: dict[str, list[str]],
    exclude_null_score_security: bool = True,
    show_all_data: bool = False,
    hold_securities: Optional[list[str]] = None,
) -> pl.LazyFrame:
    """规划选标计算"""
    # 规划选标截面
    screen_lf = screen_lf.pipe(
        planning_screen_lf,
        parameters=parameters,
        builtin_factors=builtin_factors,
        filter_exprs=filter_exprs,
        BuiltinScopes=BuiltinScopes,
        ProcessMissingDataMethods=ProcessMissingDataMethods,
        NeutralizationMethods=NeutralizationMethods,
        filter_fn=lambda x: x.filter(pl.col("trade_date") == trade_date),
    )
    # 添加权重列
    screen_lf = screen_lf.pipe(planning_screen_weight, parameters["hold_method"])
    # 若阈值不为0 且 存在持仓数据 => 规划计算修正排名
    if parameters["hold_method"]["hold_threshold"] and hold_securities:
        position_lf = pl.LazyFrame(
            {"code": hold_securities}, schema={"code": pl.String}
        ).with_columns(pl.lit(1).alias("volume"))
        screen_lf = screen_lf.pipe(
            planning_mod_rank,
            parameters["hold_method"]["hold_threshold"],
            parameters["hold_method"]["hold_range_type"],
            position_lf,
        )
    # 排除得分null的证券
    if exclude_null_score_security:
        screen_lf = screen_lf.filter(pl.col("score").is_not_null())
    screen_lf = (
        screen_lf
        # 按照rank排序
        .sort("rank", descending=False, nulls_last=True)
        # 添加选中列
        .with_columns(
            pl.col("rank")
            .is_between(
                parameters["hold_method"]["hold_range_value"]["start"],
                parameters["hold_method"]["hold_range_value"]["end"],
            )
            .fill_null(False)
            .alias("selected"),
            pl.col("time").cast(pl.String).alias("time"),  # 转换 time 为字符串
        )
        # 生成返回数据列
        .select(
            _gen_screen_result_columns(
                fixed_columns, parameters, setting_relative_factors, show_all_data
            )
        )
        # 调整返回列顺序
        .select(
            cs.by_name([c for c in fixed_columns]),  # fixed_columns 在前
            cs.exclude(fixed_columns),  # 其他列在后
        )
    )
    return screen_lf


def cb_screen(
    screen_lf: pl.LazyFrame,
    trade_date: date,
    parameters_list: list[CBScreenParameters],
    exclude_null_score_security: bool = True,
    show_all_data: bool = False,
    hold_securities: Optional[list[str]] = None,
) -> list[pl.DataFrame]:
    """选债"""
    # 生成选债计划
    screen_plans: list[pl.LazyFrame] = [
        _planning_screen(
            screen_lf=screen_lf,
            trade_date=trade_date,
            parameters=parameters,
            builtin_factors=CB_BUILTIN_FACTORS,
            BuiltinScopes=CBBuiltinScope,
            ProcessMissingDataMethods=CBProcessMissingDataMethod,
            NeutralizationMethods=CBNeutralizationMethod,
            filter_exprs=gen_cb_filter_exprs(parameters),
            fixed_columns=CB_SCREEN_RESULT_FIXED_COLUMNS,
            setting_relative_factors=SCREEN_CB_SETTING_RELATIVE_FACTORS,
            exclude_null_score_security=exclude_null_score_security,
            show_all_data=show_all_data,
            hold_securities=hold_securities,
        )
        for parameters in parameters_list
    ]
    # 执行选债计划
    return pl.collect_all(screen_plans)


def stk_screen(
    screen_lf: pl.LazyFrame,
    trade_date: date,
    parameters_list: list[STKScreenParameters],
    exclude_null_score_security: bool = True,
    show_all_data: bool = False,
    hold_securities: Optional[list[str]] = None,
):
    """选股"""
    # 生成选股计划
    screen_plans: list[pl.LazyFrame] = [
        _planning_screen(
            screen_lf=screen_lf,
            trade_date=trade_date,
            parameters=parameters,
            builtin_factors=STK_BUILTIN_FACTORS,
            BuiltinScopes=STKBuiltinScope,
            ProcessMissingDataMethods=STKProcessMissingDataMethod,
            NeutralizationMethods=STKNeutralizationMethod,
            filter_exprs=gen_stk_filter_exprs(parameters),
            fixed_columns=STK_SCREEN_RESULT_FIXED_COLUMNS,
            setting_relative_factors=SCREEN_STK_SETTING_RELATIVE_FACTORS,
            exclude_null_score_security=exclude_null_score_security,
            show_all_data=show_all_data,
            hold_securities=hold_securities,
        )
        for parameters in parameters_list
    ]
    # 执行选股计划
    return pl.collect_all(screen_plans)


def etf_screen(
    screen_lf: pl.LazyFrame,
    trade_date: date,
    parameters_list: list[ETFScreenParameters],
    exclude_null_score_security: bool = True,
    show_all_data: bool = False,
    hold_securities: Optional[list[str]] = None,
):
    """选基"""
    # 历史数据预切片：取所有策略中最大的时间窗口，只计算一次 cutoff
    max_window = max(
        (extract_max_window(p.get("computed_factors")) for p in parameters_list),
        default=0,
    )
    if max_window > 0:
        screen_lf = slice_history_by_window(screen_lf, max_window)
    # 生成选基计划
    screen_plans: list[pl.LazyFrame] = [
        _planning_screen(
            screen_lf=screen_lf,
            trade_date=trade_date,
            parameters=parameters,
            builtin_factors=ETF_BUILTIN_FACTORS,
            BuiltinScopes=ETFBuiltinScope,
            ProcessMissingDataMethods=ETFProcessMissingDataMethod,
            NeutralizationMethods=ETFNeutralizationMethod,
            filter_exprs=gen_etf_filter_exprs(parameters),
            fixed_columns=ETF_SCREEN_RESULT_FIXED_COLUMNS,
            setting_relative_factors=SCREEN_ETF_SETTING_RELATIVE_FACTORS,
            exclude_null_score_security=exclude_null_score_security,
            show_all_data=show_all_data,
            hold_securities=hold_securities,
        )
        for parameters in parameters_list
    ]
    # 执行选基计划
    return pl.collect_all(screen_plans)



from datetime import date, datetime
from typing import Type

import polars as pl
from polars_event_backtest import event_backtest as rs_event_backtest

from ldnode import logger
from ldnode.calculator.plans.polars import (
    planning_mod_rank,
    planning_screen_lf,
)
from ldnode.calculator.schemas import (
    CB_BUILTIN_FACTORS,
    EPSILON,
    ETF_BUILTIN_FACTORS,
    STK_BUILTIN_FACTORS,
    BacktestParameters,
    BuiltinScope,
    CBBacktestParameters,
    CBBuiltinScope,
    CBNeutralizationMethod,
    CBProcessMissingDataMethod,
    ETFBacktestParameters,
    ETFBuiltinScope,
    ETFNeutralizationMethod,
    ETFProcessMissingDataMethod,
    HistoryPosition,
    HoldWeight,
    NeutralizationMethod,
    Portfolio,
    ProcessMissingDataMethod,
    STKBacktestParameters,
    STKBuiltinScope,
    STKNeutralizationMethod,
    STKProcessMissingDataMethod,
    StopType,
    gen_cb_filter_exprs,
    gen_etf_filter_exprs,
    gen_stk_filter_exprs,
)

from .calc_rebalance_days import calc_rebalance_days


def _rs_event_backtest(
    lf: pl.LazyFrame,
    parameters: BacktestParameters,
    builtin_factors: list[str],
    BuiltinScopes: Type[BuiltinScope],
    ProcessMissingDataMethods: Type[ProcessMissingDataMethod],
    NeutralizationMethods: Type[NeutralizationMethod],
    board_lot: int,
    filter_exprs: list[pl.Expr] = [],
    record_history: bool = True,
    # Issue #1：新增参数，控制是否在返回值中附带 planning_screen_lf 全列截面数据
    # factor_analysis handler 需要全因子列来计算 Rank IC
    return_screen_df: bool = False,
) -> tuple[pl.DataFrame, pl.DataFrame] | tuple[pl.DataFrame, pl.DataFrame, pl.DataFrame]:
    """事件驱动回测(Rust实现)"""
    # 交易日范围series
    trade_dates_df = (
        lf.filter(
            pl.col("trade_date").is_between(
                lower_bound=datetime.strptime(
                    parameters["date_range"]["start_date"], "%Y-%m-%d"
                ),
                upper_bound=datetime.strptime(
                    parameters["date_range"]["end_date"], "%Y-%m-%d"
                ),
            )
        )
        .select("trade_date")
        .unique()
        .sort("trade_date")
        .set_sorted("trade_date")  # 将列标记为已排序,加快未来的操作
    ).collect()
    filter_exprs.extend([pl.col("suspension"), pl.col("last_trade_day")])
    lf = (
        lf.pipe(
            planning_screen_lf,
            parameters=parameters,
            builtin_factors=builtin_factors,
            filter_exprs=filter_exprs,
            BuiltinScopes=BuiltinScopes,
            ProcessMissingDataMethods=ProcessMissingDataMethods,
            NeutralizationMethods=NeutralizationMethods,
            filter_fn=lambda x: x.filter(
                pl.col("trade_date").is_between(
                    lower_bound=datetime.strptime(
                        parameters["date_range"]["start_date"], "%Y-%m-%d"
                    ),
                    upper_bound=datetime.strptime(
                        parameters["date_range"]["end_date"], "%Y-%m-%d"
                    ),
                )
            ),
        ).sort(["trade_date", "rank"], descending=False)  # 同交易日内按rank排序
    )
    rebalance_days = calc_rebalance_days(
        frame=trade_dates_df.with_row_index(), parameters=parameters
    )
    # 事件循环所需列
    event_loop_need_columns = [
        "code",
        "name",
        "trade_date",
        "close",
        "stk_div",
        "cash_div",
        "rank",
        "limit",
        "suspension",
    ]
    # 添加持仓权重所需列
    if parameters["hold_method"]["hold_weight"] == HoldWeight.AMOUNT:
        event_loop_need_columns.append("amount")  # 添加成交量列
    # 添加止盈/止损所需列
    if parameters["stop_method"]["stop_type"] != StopType.NOTSTOP:
        _stop_trade_price_expr = (
            pl.col("pre_close") * (1 + parameters["stop_method"]["stop_value"])
        ).round(3)
        if parameters["stop_method"]["stop_type"] == StopType.TAKEPROFIT:
            # 若开盘价高于止盈价格，则止盈价格为开盘价
            lf = lf.with_columns(
                pl.when(pl.col("high") > _stop_trade_price_expr)
                .then(True)
                .otherwise(False)
                .alias("_stop_trade_flag"),
                pl.max_horizontal(_stop_trade_price_expr, pl.col("open")).alias(
                    "_stop_trade_price"
                ),
            )
        elif parameters["stop_method"]["stop_type"] == StopType.STOPLOSS:
            # 若开盘价低于止损价格，则止损价格为开盘价
            lf = lf.with_columns(
                pl.when(pl.col("low") < _stop_trade_price_expr)
                .then(True)
                .otherwise(False)
                .alias("_stop_trade_flag"),
                pl.min_horizontal(_stop_trade_price_expr, pl.col("open")).alias(
                    "_stop_trade_price"
                ),
            )
        event_loop_need_columns.extend(
            ["_stop_trade_price", "_stop_trade_flag"]
        )  # 添加止盈/止损价格列
    # factor_analysis 需要全因子列计算 Rank IC，其它场景只需事件循环列
    if return_screen_df:
        screen_df = lf.collect()
        event_df = screen_df.select(event_loop_need_columns)
    else:
        screen_df = None
        event_df = lf.select(event_loop_need_columns).collect()
    result = rs_event_backtest(
        capital=parameters["capital"],
        commission=parameters["commission"],
        board_lot=board_lot,
        df=event_df,
        rebalance_days_df=rebalance_days,
        record_history=record_history,
        stop_type=parameters["stop_method"]["stop_type"],
        hold_range_type=parameters["hold_method"]["hold_range_type"],
        hold_range_start=parameters["hold_method"]["hold_range_value"]["start"],
        hold_range_end=parameters["hold_method"]["hold_range_value"]["end"],
        hold_weight_type=parameters["hold_method"]["hold_weight"],
        max_hold_weight=parameters["hold_method"]["max_hold_weight"],
        hold_threshold=parameters["hold_method"]["hold_threshold"],
        hold_weight_factor=None,
    )
    if return_screen_df:
        return result[0], result[1], screen_df
    return result[0], result[1]


def _event_backtest(
    lf: pl.LazyFrame,
    parameters: BacktestParameters,
    builtin_factors: list[str],
    BuiltinScopes: Type[BuiltinScope],
    ProcessMissingDataMethods: Type[ProcessMissingDataMethod],
    NeutralizationMethods: Type[NeutralizationMethod],
    board_lot: int,
    filter_exprs: list[pl.Expr] = [],
    record_history: bool = True,
) -> tuple[pl.DataFrame, dict[str, HistoryPosition]]:
    """事件驱动回测"""
    equity_list = []
    turnover_list = []
    commission_list = []
    hold_securities_list = []

    # 交易日范围series
    trade_dates_df = (
        lf.filter(
            pl.col("trade_date").is_between(
                lower_bound=datetime.strptime(
                    parameters["date_range"]["start_date"], "%Y-%m-%d"
                ),
                upper_bound=datetime.strptime(
                    parameters["date_range"]["end_date"], "%Y-%m-%d"
                ),
            )
        )
        .select("trade_date")
        .unique()
        .sort("trade_date")
        .set_sorted("trade_date")  # 将列标记为已排序,加快未来的操作
    ).collect()

    filter_exprs.extend([pl.col("suspension"), pl.col("last_trade_day")])
    lf = (
        lf.pipe(
            planning_screen_lf,
            parameters=parameters,
            builtin_factors=builtin_factors,
            filter_exprs=filter_exprs,
            BuiltinScopes=BuiltinScopes,
            ProcessMissingDataMethods=ProcessMissingDataMethods,
            NeutralizationMethods=NeutralizationMethods,
            filter_fn=lambda x: x.filter(
                pl.col("trade_date").is_between(
                    lower_bound=datetime.strptime(
                        parameters["date_range"]["start_date"], "%Y-%m-%d"
                    ),
                    upper_bound=datetime.strptime(
                        parameters["date_range"]["end_date"], "%Y-%m-%d"
                    ),
                )
            ),
        ).sort(["trade_date", "rank"], descending=False)  # 同交易日内按rank排序
    )
    rebalance_days = calc_rebalance_days(
        frame=trade_dates_df.with_row_index(), parameters=parameters
    )
    # 初始化持仓表
    code_categories = lf.collect_schema().get("code")
    p = Portfolio(
        code_categories=code_categories,  # type: ignore
        logger=logger.event_backtest,
        balance=float(parameters["capital"]),
        commission=float(parameters["commission"]),
    )
    # logger.event_backtest.info(f"初始资金: {parameters['capital']}, 手续费率: {parameters['commission']}")
    # 事件循环所需列
    event_loop_need_columns = [
        "code",
        "name",
        "trade_date",
        "close",
        "stk_div",
        "cash_div",
        "rank",
        "limit",
        "suspension",
    ]
    # 添加持仓权重所需列
    if parameters["hold_method"]["hold_weight"] == HoldWeight.AMOUNT:
        event_loop_need_columns.append("amount")  # 添加成交量列
    # 添加止盈/止损所需列
    if parameters["stop_method"]["stop_type"] != StopType.NOTSTOP:
        _stop_trade_price_expr = (
            pl.col("pre_close") * (1 + parameters["stop_method"]["stop_value"])
        ).round(3)
        if parameters["stop_method"]["stop_type"] == StopType.TAKEPROFIT:
            # 若开盘价高于止盈价格，则止盈价格为开盘价
            lf = lf.with_columns(
                pl.when(pl.col("high") > _stop_trade_price_expr)
                .then(True)
                .otherwise(False)
                .alias("_stop_trade_flag"),
                pl.max_horizontal(_stop_trade_price_expr, pl.col("open")).alias(
                    "_stop_trade_price"
                ),
            )
        elif parameters["stop_method"]["stop_type"] == StopType.STOPLOSS:
            # 若开盘价低于止损价格，则止损价格为开盘价
            lf = lf.with_columns(
                pl.when(pl.col("low") < _stop_trade_price_expr)
                .then(True)
                .otherwise(False)
                .alias("_stop_trade_flag"),
                pl.min_horizontal(_stop_trade_price_expr, pl.col("open")).alias(
                    "_stop_trade_price"
                ),
            )
        event_loop_need_columns.extend(
            ["_stop_trade_price", "_stop_trade_flag"]
        )  # 添加止盈/止损价格列
    # logger.event_backtest.debug(f"事件循环所需列: {event_loop_need_columns}")
    # 循环
    loop_iter = (
        lf.select(event_loop_need_columns)
        .collect()
        .partition_by(
            "trade_date", maintain_order=True, include_key=False, as_dict=True
        )
        .items()
    )
    for (trade_date,), df in loop_iter:
        trade_date: date
        df: pl.DataFrame

        # 初始化计数器
        buy_amount: float = 0.0  # 本周买入金额
        sell_amount: float = 0.0  # 本周卖出金额
        buy_commission: float = 0.0  # 本周买入手续费
        sell_commission: float = 0.0  # 本周卖出手续费
        # 记录本周期开始时equity
        start_equity: float = p.equity
        # 记录本周期开始时持仓标的
        hold_securities = p.position.get_column("code")

        # 处理分红的标的
        div_df = (
            df.lazy()
            .filter(pl.col("cash_div").is_not_null() | pl.col("stk_div").is_not_null())
            .join(p.position.lazy().select("code", "volume"), on="code", how="inner")
            .select(
                pl.col("code"),
                pl.col("volume"),
                pl.col("cash_div").fill_null(0.0).alias("cash_div"),
                pl.col("stk_div").fill_null(0.0).alias("stk_div"),
            )
            .collect()
        )

        if div_df.height:
            # logger_daily.debug(f"{trade_date} 处理分红: {div_df.height}个标的")
            # 处理现金分红（扣除20%所得税）
            p.balance += (
                div_df.select(pl.col("cash_div") * pl.col("volume") / 10).sum().item()
                * 0.8
            )

            # 处理股票分红
            p.position = p.position.update(
                div_df.select(
                    pl.col("code"),
                    (
                        pl.col("volume") + (pl.col("stk_div") * pl.col("volume") / 10)
                    ).alias("volume"),
                ),
                on="code",
                how="left",
            )

            if record_history:
                # 记录分红标的
                p.history_dividend[trade_date.strftime("%Y-%m-%d")] = div_df

        # 止盈/止损模块
        if parameters["stop_method"]["stop_type"] != StopType.NOTSTOP:
            # 触发止盈/止损标的 和 持仓中的标的 的交集
            stop_trade_codes = df.filter(pl.col("_stop_trade_flag"))["code"]
            if_trigger_stop_trade = stop_trade_codes.is_in(p.position["code"]).any()
            # 若持仓中有标的触发止盈/止损
            if if_trigger_stop_trade:
                # logger_daily.debug(f"{trade_date} 触发止{'盈' if strategy_config['stop_method']['stop_type'] == StopType.TAKEPROFIT else '损'}: {stop_trade_codes.to_list()}")
                # 刷新持仓表中止盈/止损价格
                p.position = p.position.update(
                    df.filter(pl.col("_stop_trade_flag")).select(
                        pl.col("code"),
                        pl.col("_stop_trade_price").alias("last_price"),
                        pl.col("suspension"),
                    ),
                    on="code",
                    how="left",
                )
                # 用持仓中的标的生成target表
                _stop_trade_target_df = p.position.select(
                    pl.col("code"),
                    pl.col("name"),
                    pl.col("last_price"),
                    pl.col("volume"),
                    pl.when(pl.col("code").is_in(stop_trade_codes))
                    .then(pl.lit(0.0))
                    .otherwise(pl.col("volume"))
                    .alias("target_volume"),
                )
                stop_trade_target_lf = _stop_trade_target_df.select(
                    pl.exclude("volume")
                ).lazy()
                _buy_amount, _sell_amount, _buy_commission, _sell_commission = (
                    p.fill_target(stop_trade_target_lf)
                )

                buy_amount += _buy_amount  # 叠加买入金额
                sell_amount += _sell_amount  # 叠加卖出金额
                buy_commission += _buy_commission  # 叠加买入手续费
                sell_commission += _sell_commission  # 叠加卖出手续费

        # 刷新持仓表中收盘价格并更新市值
        p.position = p.position.update(
            df.select(
                pl.col("code"),
                pl.col("name"),
                pl.col("close").alias("last_price"),
                pl.col("suspension"),
            ),
            on="code",
            how="left",
        ).with_columns(
            (pl.col("last_price") * pl.col("volume")).round(2).alias("market_value")
        )

        # 判断本周期是否需要调仓
        if_rebalance = (
            rebalance_days.filter(pl.col("trade_date") == trade_date)
            .select("if_rebalance")
            .item()
        )
        if if_rebalance:
            # logger_daily.debug(f"{trade_date} 执行调仓")
            # 根据阈值修改rank
            if parameters["hold_method"]["hold_threshold"]:
                df = planning_mod_rank(
                    frame=df,
                    threshold=parameters["hold_method"]["hold_threshold"],
                    hold_range_type=parameters["hold_method"]["hold_range_type"],
                    position_lf=p.position.lazy(),
                ).collect()

            # 保留持仓范围内数据(初始化交易表)
            hold_range = parameters["hold_method"]["hold_range_value"]

            target_df = (
                df.lazy()
                .join(p.position.lazy().select("code", "volume"), on="code", how="left")
                .filter(
                    # 选出排名范围内的标的
                    pl.col("rank").is_between(
                        lower_bound=hold_range["start"],
                        upper_bound=hold_range["end"],
                        closed="both",
                    ),
                    # 排除涨停且没有持仓的标的
                    ~((pl.col("limit") == 1) & (pl.col("volume") == 0)),
                )
                .select(pl.exclude("volume"))
                .collect()
            )

            # 进行调仓
            # if target_df.height:
            # logger_daily.debug(f"目标持仓数量: {target_df.height}")
            # 计算目标持仓权重
            if parameters["hold_method"]["hold_weight"] == HoldWeight.AVG:
                _weight_expr = pl.lit(1) / target_df.height  # 等权重
            elif parameters["hold_method"]["hold_weight"] == HoldWeight.CLOSE:
                _weight_expr = pl.col("close") / pl.col("close").sum()  # 按价格加权
            elif parameters["hold_method"]["hold_weight"] == HoldWeight.AMOUNT:
                _weight_expr = pl.col("amount") / pl.col("amount").sum()  # 按成交额加权
            elif parameters["hold_method"]["hold_weight"] == HoldWeight.NOTHOLD:
                _weight_expr = pl.lit(0.0)
            else:
                raise ValueError(
                    f"Invalid hold weight: {parameters['hold_method']['hold_weight']}"
                )

            # 计算目标持仓数量(预扣除手续费后再向下取整)
            target_lf = (
                target_df.lazy()
                .select(
                    pl.col("code"),
                    pl.col("name"),
                    pl.col("close"),
                    (
                        (
                            _weight_expr.clip(
                                upper_bound=parameters["hold_method"]["max_hold_weight"]
                            )
                            * p.available_equity
                            * (1 - p.commission) ** 2
                            // (pl.col("close") * board_lot)
                        )
                        * board_lot
                    ).alias("target_volume"),
                )
                .rename({"close": "last_price"})
            )
            _buy_amount, _sell_amount, _buy_commission, _sell_commission = (
                p.fill_target(target_lf)
            )

            buy_amount += _buy_amount  # 叠加买入金额
            sell_amount += _sell_amount  # 叠加卖出金额
            buy_commission += _buy_commission  # 叠加买入手续费
            sell_commission += _sell_commission  # 叠加卖出手续费

        # 非调仓日 且触发止盈/止损 买回标的
        if (
            not if_rebalance
            and parameters["stop_method"]["stop_type"] != StopType.NOTSTOP
            and if_trigger_stop_trade
        ):
            # logger_daily.info(f"{trade_date} 非调仓日买回止{'盈' if strategy_config['stop_method']['stop_type'] == StopType.TAKEPROFIT else '损'}标的")
            # 买回止盈/止损标的
            stop_trade_target_lf = (
                _stop_trade_target_df.select(pl.exclude("target_volume"))
                .rename({"volume": "target_volume"})
                .update(
                    df.select(
                        pl.col("code"),
                        pl.col("close").alias("last_price"),  # 用收盘价填充
                    ),
                    on="code",
                    how="left",
                )
            ).lazy()
            _buy_amount, _sell_amount, _buy_commission, _sell_commission = (
                p.fill_target(stop_trade_target_lf)
            )

            buy_amount += _buy_amount  # 叠加买入金额
            sell_amount += _sell_amount  # 叠加卖出金额
            buy_commission += _buy_commission  # 叠加买入手续费
            sell_commission += _sell_commission  # 叠加卖出手续费

        if record_history:
            p.update_history_position(trade_date=trade_date.strftime("%Y-%m-%d"))

        # 保存equity, turnover, commission, hold_securities
        turnover_list.append(
            (buy_amount + sell_amount) / 2 / start_equity if start_equity else EPSILON
        )
        equity_list.append(p.equity)
        commission_list.append(buy_commission + sell_commission)
        hold_securities_list.append(hold_securities)

    # 生成res
    res = (
        trade_dates_df.lazy()
        .with_columns(
            pl.Series(name="equity", values=equity_list),
            pl.Series(name="turnover", values=turnover_list),
            pl.Series(name="commission", values=commission_list),
            pl.Series(name="hold_securities", values=hold_securities_list),
        )
        .with_columns(
            pl.col("equity")
            .pct_change()
            .fill_null(equity_list[0] / parameters["capital"] - 1)
            .alias("time_return"),  # 计算每周期回报率
        )
        .collect()
    )

    # logger.event_backtest.info(f"回测完成，最终净值: {equity_list[-1] / parameters['capital']:.4f}")
    # logger.event_backtest_daily.info(f"回测完成，最终净值: {equity_list[-1] / parameters['capital']:.4f}")
    return res, p.history_position


def cb_event_backtest(
    lf: pl.LazyFrame,
    parameters: CBBacktestParameters,
    record_history: bool = True,
    # Issue #1：透传给 _rs_event_backtest，控制是否返回全列截面数据
    return_screen_df: bool = False,
) -> tuple[pl.DataFrame, pl.DataFrame] | tuple[pl.DataFrame, pl.DataFrame, pl.DataFrame]:
    """转债事件驱动回测"""
    filter_exprs = gen_cb_filter_exprs(parameters)
    return _rs_event_backtest(
        lf=lf,
        parameters=parameters,
        builtin_factors=CB_BUILTIN_FACTORS,
        BuiltinScopes=CBBuiltinScope,
        ProcessMissingDataMethods=CBProcessMissingDataMethod,
        NeutralizationMethods=CBNeutralizationMethod,
        board_lot=10,
        filter_exprs=filter_exprs,
        record_history=record_history,
        return_screen_df=return_screen_df,
    )


def stk_event_backtest(
    lf: pl.LazyFrame,
    parameters: STKBacktestParameters,
    record_history: bool = True,
    # Issue #1：透传给 _rs_event_backtest，控制是否返回全列截面数据
    return_screen_df: bool = False,
) -> tuple[pl.DataFrame, pl.DataFrame] | tuple[pl.DataFrame, pl.DataFrame, pl.DataFrame]:
    """股票事件驱动回测"""
    filter_exprs = gen_stk_filter_exprs(parameters)
    return _rs_event_backtest(
        lf=lf,
        parameters=parameters,
        builtin_factors=STK_BUILTIN_FACTORS,
        BuiltinScopes=STKBuiltinScope,
        ProcessMissingDataMethods=STKProcessMissingDataMethod,
        NeutralizationMethods=STKNeutralizationMethod,
        board_lot=100,
        filter_exprs=filter_exprs,
        record_history=record_history,
        return_screen_df=return_screen_df,
    )


def etf_event_backtest(
    lf: pl.LazyFrame,
    parameters: ETFBacktestParameters,
    record_history: bool = True,
    # Issue #1：透传给 _rs_event_backtest，控制是否返回全列截面数据
    return_screen_df: bool = False,
) -> tuple[pl.DataFrame, pl.DataFrame] | tuple[pl.DataFrame, pl.DataFrame, pl.DataFrame]:
    """基金事件驱动回测"""
    filter_exprs = gen_etf_filter_exprs(parameters)
    return _rs_event_backtest(
        lf=lf,
        parameters=parameters,
        builtin_factors=ETF_BUILTIN_FACTORS,
        BuiltinScopes=ETFBuiltinScope,
        ProcessMissingDataMethods=ETFProcessMissingDataMethod,
        NeutralizationMethods=ETFNeutralizationMethod,
        board_lot=100,
        filter_exprs=filter_exprs,
        record_history=record_history,
        return_screen_df=return_screen_df,
    )



from typing import Optional

import polars as pl

from ldnode.calculator.schemas import SortOrder


def planning_rank_factors(
    frame: pl.LazyFrame | pl.DataFrame,
    pct: bool = False,
    more_by: Optional[list[tuple[str, SortOrder]]] = None,
) -> pl.LazyFrame:
    """ 规划因子排名计算 """
    lf = frame.lazy() if isinstance(frame, pl.DataFrame) else frame
    # 处理more_by默认值
    more_by = more_by or [('close', SortOrder.ASCENDING)]
    # 构建排序表达式列表 - 确保score在前面
    sort_columns = ['trade_date', 'score']  # 首先按交易日期，然后按score
    sort_flags = [False, True]  # trade_date 升序, score 降序
    # 添加more_by字段的表达式
    for field, sort_order in more_by:
        sort_columns.append(field)
        sort_flags.append(sort_order == SortOrder.DESCENDING)
    # 计算排名表达式
    if pct:
        rank_expr = (
            pl.col('score')
            .rank(method='average', descending=True)
            .over('trade_date')
            .truediv(
                pl.col('score').count().over('trade_date')
            )
        )
    else:
        rank_expr = (
            pl.col('score')
            .rank(method='ordinal', descending=True)
            .over('trade_date')
        )
    return (
        lf
        .sort(sort_columns, descending=sort_flags)  # 使用表达式列表，确保排序包含score
        .with_columns(rank_expr.alias('rank'))
    )
    


from .factor import ProcessMissingDataMethod, NeutralizationMethod
from .screen import BuiltinScope


class ETFNeutralizationMethod(NeutralizationMethod):
    """ ETF中性化方法 """


class ETFProcessMissingDataMethod(ProcessMissingDataMethod):
    """ ETF缺失值处理方式 """


class ETFBuiltinScope(BuiltinScope):
    """基金内置交易所板块"""



from enum import IntEnum
from typing import TypedDict, Any


class InputMessageType(IntEnum):
    """ 通道接受消息类型 """
    TASK_SUBMITTED = 0 # 任务提交
    TASK_CANCELLED = 1 # 任务取消
    INVOKE_ANSWER = 2 # 调用应答


class OutputMessageType(IntEnum):
    """ 通道发送消息类型 """
    TASK_ANSWER = 0 # 任务应答
    INVOKE = 1 # 调用
    EVENT = 2 # 事件


class InputMessage(TypedDict):
    """ 输入消息 """
    msg_type: InputMessageType # 消息类型
    payload: Any # 消息负载
    
    
class OutputMessage(TypedDict):
    """ 输出消息 """
    msg_type: OutputMessageType # 消息类型
    payload: Any # 消息负载


class InvokeResult(TypedDict):
    """ 调用应答 """
    task_id: str # 任务ID
    result: Any # 结果



import time
from functools import wraps
from logging import Logger
from typing import Any, Callable, Generator

import humanize

from ldnode import logger
from ldnode.calculator.handlers.models import (
    TaskBaseParameters,
    TaskResponse,
)

from .responses import error_response, response


def wrap_handler(
    task_name: str, logger: Logger = logger.calculator, is_generator: bool = False
):
    """
    任务处理装饰器
    运行时间统计、异常处理、日志记录
    """
    if is_generator:

        def flow_decorator(
            handler: Callable[..., Generator[TaskResponse[Any]]],
        ):
            @wraps(handler)
            def wrapper(
                parameters: TaskBaseParameters, **kwargs
            ) -> Generator[TaskResponse[Any]]:
                logger.info(f"{task_name}({parameters['task_id']}) 开始执行")
                start_time = time.time()
                start_cpu_time = time.process_time()
                generator = handler(parameters, **kwargs)
                try:
                    for result in generator:
                        yield result
                    logger.info(f"{task_name}({parameters['task_id']}) 执行成功")
                except Exception as e:
                    logger.exception(
                        f"{task_name}({parameters['task_id']}) 执行失败: {e}"
                    )
                    logger.error(
                        f"{task_name}({parameters['task_id']}) 执行参数: {parameters}"
                    )
                    yield error_response(e)
                finally:
                    run_time = time.time() - start_time
                    cpu_time = time.process_time() - start_cpu_time
                    logger.info(
                        f"{task_name}({parameters['task_id']}) "
                        f"执行时间: {humanize.naturaldelta(run_time, minimum_unit='milliseconds')}({run_time * 1000}ms) "
                        f"(CPU: {humanize.naturaldelta(cpu_time, minimum_unit='milliseconds')})"
                    )
                    yield response("fee", {"run_time": run_time, "cpu_time": cpu_time})

            return wrapper

        return flow_decorator
    else:

        def decorator(handler: Callable[..., TaskResponse[Any]]):
            @wraps(handler)
            def wrapper(parameters: TaskBaseParameters, **kwargs) -> TaskResponse[Any]:
                start_time = time.time()
                start_cpu_time = time.process_time()
                try:
                    logger.info(f"{task_name}({parameters['task_id']}) 开始执行")
                    result = handler(parameters, **kwargs)
                    logger.info(f"{task_name}({parameters['task_id']}) 执行成功")
                except Exception as e:
                    logger.exception(
                        f"{task_name}({parameters['task_id']}) 执行失败: {e}"
                    )
                    logger.error(
                        f"{task_name}({parameters['task_id']}) 执行参数: {parameters}"
                    )
                    result = error_response(e)
                finally:
                    run_time = time.time() - start_time
                    cpu_time = time.process_time() - start_cpu_time
                    logger.info(
                        f"{task_name}({parameters['task_id']}) "
                        f"执行时间: {humanize.naturaldelta(run_time, minimum_unit='milliseconds')}({run_time * 1000}ms) "
                        f"(CPU: {humanize.naturaldelta(cpu_time, minimum_unit='milliseconds')})"
                    )
                    result.update(
                        {
                            "run_time": run_time,
                            "cpu_time": cpu_time,
                        }
                    )
                    return result

            return wrapper

        return decorator



import polars as pl


def plan_rename(lf: pl.LazyFrame) -> pl.LazyFrame:
    return lf.rename({
        'close': 'pre_close',
        'turnover': 'vol',
        'volume': 'amount',
        'date': 'trade_date'
    })
    


import os
import signal
import subprocess
import sys
import time

import psutil

from ldnode.runner.events import Ping


def load_ping_info() -> Ping:
    """加载心跳信息"""
    return {
        "report_time": int(time.time()),
        "cpu_usage": psutil.cpu_percent(),
        "memory_usage": psutil.virtual_memory().percent,
        "disk_usage": psutil.disk_usage("/").percent,
        "load_average": psutil.getloadavg()[0],
    }


def _sanitize_fds() -> None:
    """确保 fd 0/1/2 有效——nohup 后台运行时这些 fd 可能已关闭，
    导致子进程 Python 解释器 init_sys_streams 失败"""
    for fd in (0, 1, 2):
        try:
            os.fstat(fd)
        except OSError:
            os.open(os.devnull, os.O_RDWR)  # 自动分配到最小可用 fd


def restart() -> None:
    """重启进程
    LaunchDaemon 模式：以非0退出码退出，由 launchd 自动重拉起
    非 LaunchDaemon 模式：fork + execl 自行拉起新进程
    """
    if os.environ.get("LAUNCHD_MANAGED"):
        os.kill(os.getpid(), signal.SIGTERM)
        time.sleep(5)
        # 优雅关闭超时，强制以非0退出码退出让 launchd 重启
        sys.exit(1)
    else:
        pid = os.fork()
        if pid == 0:
            _sanitize_fds()
            time.sleep(3)  # 等待3秒后重启
            python_path = sys.executable
            os.execl(python_path, python_path, *sys.argv)
        else:
            # 结束当前主进程，走既有优雅关闭流程（释放工作子进程）
            os.kill(os.getpid(), signal.SIGTERM)


def upgrade(version: str | None = None) -> tuple[bool, str]:
    """升级 ldnode"""
    package = "ldnode" if not version else f"ldnode=={version}"
    cmd = [sys.executable, "-m", "pip", "install", "-U", package]
    try:
        process = subprocess.run(
            cmd,
            stdin=subprocess.DEVNULL,
            capture_output=True,
            text=True,
            timeout=900,
            check=False,
        )
    except Exception as e:
        return False, f"执行升级命令失败: {e}"
    output = (process.stdout or "") + (process.stderr or "")
    output = output[-4096:]
    if process.returncode == 0:
        return True, output
    return False, output



import asyncio
from typing import Callable

from ldnode.runner.schemas import BaseContext
from ldnode.runner.events import Events


class LoadCounter:
    """ 负载计数器 """
    DECAY_1M = 0.92 # interval: 5s
    
    def __init__(self, context: BaseContext, load_querier: Callable[[], float], running_checker: Callable[[], bool]):
        self.context = context
        self.load_1m: float = 0
        self.tick_task: asyncio.Task | None = None
        # 外部接口
        self.query_load = load_querier
        self.check_running = running_checker
        
    def _update(self, load_now: float) -> None:
        """ 更新负载 """
        self.load_1m = self.load_1m * self.DECAY_1M + load_now * (1 - self.DECAY_1M)

    def _ewma(self) -> None:
        """ 指数加权移动平均更新负载 """
        self._update(self.query_load())
        
    async def _tick_loop(self) -> None:
        """ 定时循环 """
        if self.check_running():
            self._ewma()
            await self.context.submit_event(
                "calculator", 
                Events.UPDATE_CALCULATOR_LOAD, 
                { "load": self.load_1m },
            )
            await asyncio.sleep(5)
            self.tick_task = asyncio.create_task(self._tick_loop())
            
    async def count(self) -> None:
        """ 开始计数 """
        self.tick_task = asyncio.create_task(self._tick_loop())



from functools import lru_cache

from .settings import Settings


@lru_cache()
def get_settings() -> Settings:
    """
        获取配置
        @return: 配置
    """
    return Settings()


__all__ = ['get_settings', "Settings"]



from typing import TypedDict, Any, TypeVar, Generic

from ldnode.calculator.schemas.enumerations import FactorType, ExprType, RankScope, ProcessMissingDataMethod, NeutralizationMethod


NFT = TypeVar("NFT", bound=NeutralizationMethod)
""" 中性化因子类型 """
MMT = TypeVar("MMT", bound=ProcessMissingDataMethod)
""" 缺失值处理方式类型 """


class ExcludeFactor(TypedDict):
    """ 排除因子 """
    factor: str # 因子
    type: FactorType # 因子类型
    operator: str # 操作符
    value: Any # 值
    

class ComputedFactor(TypedDict):
    """ 自定义因子 """
    factor: str # 因子
    expr: str # 表达式
    expr_type: ExprType # 表达式类型
    is_intermediate: bool # 是否是中间过程


class RankFactor(TypedDict, Generic[NFT, MMT]):
    """ 打分因子配置 """
    factor: str # 因子名
    type: FactorType # 因子类型
    ascending: bool # 是否升序
    weight: float # 权重
    scope: RankScope # 行业级别
    neutralization_factor: list[NFT | ComputedFactor] # 中性化因子
    process_missing_data_method: MMT # 缺失值处理方式



from clickhouse_driver import Client

from ldnode.runner.config import get_settings


settings = get_settings()


def get_clickhouse_client() -> Client:
    """
        获取Clickhouse客户端

        :return: Clickhouse客户端
        :rtype: clickhouse_driver.Client
    """
    return Client.from_url(settings.CLICKHOUSE_URI)



import os
from pathlib import Path

from minio import Minio

from ldnode import logger
from ldnode.loader.schemas.enumerations import SyncStatus
from ldnode.loader.schemas.models import DataSource, SyncResult


class S3QuotationDataSource(DataSource):
    """
    S3 行情后台预拉取数据源。

    由 LoaderWorker 注册为 LoadRule（每 3s，仅交易时段），替代原本地行情计算逻辑。

    使用 MinIO 客户端全量 GET 下载整个文件，
    绕过 polars scan_parquet 的 HTTP Range Request（RustFS alpha 对 >1MB 文件的
    range read 存在 bug: read_file_stream offset+length 计算越界）。
    output_lazyframe() 继续读本地文件，选股请求无 S3 网络延迟。
    """

    def __init__(self, s3_uri: str, storage_options: dict, local_data_dir: Path):
        self.local_data_dir = local_data_dir
        opts = storage_options or {}
        # s3://bucket/key → bucket, key
        stripped = s3_uri.replace("s3://", "")
        slash = stripped.index("/")
        self._bucket = stripped[:slash]
        self._object_key = stripped[slash + 1:]
        # 构建 MinIO 客户端
        endpoint = opts.get("endpoint_url", "")
        secure = not endpoint.startswith("http://")
        # MinIO 客户端 endpoint 不带 scheme
        host = endpoint.replace("http://", "").replace("https://", "")
        self._client = Minio(
            host,
            access_key=opts.get("aws_access_key_id", ""),
            secret_key=opts.get("aws_secret_access_key", ""),
            secure=secure,
            region=opts.get("aws_region", "us-east-1"),
        )

    def load(self, key: str) -> list[SyncResult]:
        """全量 GET 下载 S3 对象并写入本地文件"""
        local_path = self.local_data_dir / key
        try:
            resp = self._client.get_object(self._bucket, self._object_key)
            try:
                data = resp.read()
            finally:
                resp.close()
                resp.release_conn()
            # 原子写：先写临时文件，再 rename 替换，避免读写竞态
            tmp_path = local_path.with_suffix(".tmp")
            tmp_path.write_bytes(data)
            os.rename(tmp_path, local_path)
            logger.loader.debug(f"S3 行情已同步: {key}  size={len(data)}")
            return [SyncResult(key, SyncStatus.SUCCESS, f"size={len(data)}")]
        except Exception as e:
            logger.loader.warning(f"S3 行情同步失败: {key}  error={e}")
            return [SyncResult(key, SyncStatus.FAILED, str(e))]



import polars as pl

from ldnode.calculator.schemas import HoldWeight, BacktestParameters


def planning_weight_column(
    frame: pl.LazyFrame | pl.DataFrame,
    parameters: BacktestParameters,
    with_exclude: bool = False,
) -> pl.LazyFrame:
    """ 规划持仓权重 """
    lf = frame.lazy() if isinstance(frame, pl.DataFrame) else frame
    # 选出排名内的标的
    _between_rank_expr = pl.col('rank').is_between(
        lower_bound=parameters['hold_method']['hold_range_value']['start'],
        upper_bound=parameters['hold_method']['hold_range_value']['end']
    )
    # 根据持仓方式计算权重
    if parameters['hold_method']['hold_weight'] == HoldWeight.AVG:
        # 等权持有
        weight_expr = (
            pl.when(_between_rank_expr).then(
                pl.lit(1).truediv(
                    pl.col('code').filter(_between_rank_expr).len().over('trade_date')
                )
            ).otherwise(None)
        ) if with_exclude else (
            pl.lit(1).truediv(
                pl.len().over('trade_date')
            )
        )
    elif parameters['hold_method']['hold_weight'] == HoldWeight.CLOSE:
        # 按价格比例持有
        weight_expr = (
            pl.when(_between_rank_expr).then(
                pl.col('close').truediv(
                    pl.col('close').filter(_between_rank_expr).sum().over('trade_date')
                )
            ).otherwise(None)
        ) if with_exclude else (
            pl.col('close').truediv(
                pl.col('close').sum().over('trade_date')
            )
        )
    elif parameters['hold_method']['hold_weight'] == HoldWeight.AMOUNT:
        # 按成交额比例持有
        weight_expr = (
            pl.when(_between_rank_expr).then(
                pl.col('amount').truediv(
                    pl.col('amount').filter(_between_rank_expr).sum().over('trade_date')
                ) 
            ).otherwise(None)
        ) if with_exclude else (
            pl.col('amount').truediv(
                pl.col('amount').sum().over('trade_date')
            )
        )

    elif parameters['hold_method']['hold_weight'] == HoldWeight.NOTHOLD:
        weight_expr = pl.lit(.0)
    else:
        raise ValueError(f"Invalid hold weight: {parameters['hold_method']['hold_weight']}")
    
    # 添加权重列（向量回测专用，列名为 weight，与 vector_backtest.py 保持一致）
    return (
        lf.with_columns(
            weight_expr.alias('weight'),
        )
    )



import os

from pydantic import Field
from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    """配置"""

    # 节点信息
    NODE_VERSION: str = Field(default="", description="节点版本")
    PLATFORM_TAG: str = Field(default="", description="平台标签")

    # 端点连接配置
    ENDPOINT_URL: str = Field(description="计算端点连接地址")
    ENDPOINT_ACCESS_ID: str = Field(description="计算端点访问ID")
    ENDPOINT_ACCESS_TOKEN: str = Field(description="计算端点访问令牌")
    PING_INTERVAL: int = Field(default=5, description="心跳间隔时间(秒)")

    # 计算配置
    CPU_RESOURCE: int = Field(
        default_factory=lambda: os.cpu_count() or 1, description="CPU资源"
    )
    POLARS_MAX_THREADS: int = Field(default=1, description="Polars计算最大线程")

    # 数据同步配置
    DATA_SYNC_LOCAL_DATA_DIR: str = Field(
        default="./data", description="数据同步本地数据目录"
    )
    # 数据同步COS配置
    DATA_SYNC_COS_REGION: str = Field(default="ap-shanghai", description="COS区域")
    DATA_SYNC_COS_SECRET_ID: str = Field(default="", description="COS Secret ID")
    DATA_SYNC_COS_SECRET_KEY: str = Field(default="", description="COS Secret Key")
    DATA_SYNC_COS_BUCKET: str = Field(default="", description="COS Bucket")
    # 结果数据COS配置
    RESULT_COS_REGION: str = Field(default="ap-shanghai", description="COS区域")
    RESULT_COS_SECRET_ID: str = Field(default="", description="COS Secret ID")
    RESULT_COS_SECRET_KEY: str = Field(default="", description="COS Secret Key")
    RESULT_COS_BUCKET: str = Field(default="", description="COS Bucket")

    # Clickhouse配置
    CLICKHOUSE_URI: str = Field(default="", description="Clickhouse连接URI")

    # 实时行情 S3 共享读取配置（配置后节点直接从 S3 读取，不再本地计算）
    CB_QUOTATION_S3_URI: str = Field(default="", description="转债实时行情 S3 URI（空=本地计算）")
    STK_QUOTATION_S3_URI: str = Field(default="", description="股票实时行情 S3 URI（空=本地计算）")
    ETF_QUOTATION_S3_URI: str = Field(default="", description="ETF 实时行情 S3 URI（空=本地计算）")
    QUOTATION_S3_KEY_ID: str = Field(default="", description="行情 S3 Access Key ID")
    QUOTATION_S3_SECRET_KEY: str = Field(default="", description="行情 S3 Secret Access Key")
    QUOTATION_S3_ENDPOINT_URL: str = Field(default="", description="行情 S3 Endpoint URL（私有部署时需填）")

    class Config:
        """Pydantic config for settings."""

        env_file = ".env"
        env_file_encoding = "utf-8"
        case_sensitive = True
        extra = "allow"



from .schemas.models import File

TRADE_DAYS = File(key="dim/trade_days.parquet")  # 交易日历
INDEX_DATA = File(key="backtest_data/index.parquet")  # 回测基准指数

CB_DATA = File(key="cb_data.pq")  # 转债回测数据
CB_PRE_QUOTATION = File(key="before_trading/pre_quotations.pq")  # 转债盘前数据
CB_HIST_DATA = File(key="before_trading/hist_data.pq")  # 转债历史行情
CB_QUOTATION = File(key="cb_quotations.pq")  # 转债实时截面

STK_DATA = File(key="stk_data.pq")  # 股票回测数据
STK_PRE_QUOTATION = File(key="before_trading/stk_pre_quotations.pq")  # 股票盘前数据
STK_HIST_DATA = File(key="before_trading/hist_data_stk.pq")  # 股票历史行情
STK_QUOTATION = File(key="stk_quotations.pq")  # 股票实时截面

ETF_DATA = File(key="backtest_data/etf.parquet")  # ETF回测数据
ETF_PRE_QUOTATION = File(key="before_trading/etf.parquet")  # ETF盘前数据
ETF_QUOTATION = File(key="etf_quotations.pq")  # ETF实时截面



import time
from datetime import datetime

import humanize
import easyquotation
import polars as pl
from polars import col

from ldnode import logger
from ldnode.loader.schemas.enumerations import SyncStatus
from ldnode.loader.schemas.models import DataSource, SyncResult
from ldnode.loader.files import CB_PRE_QUOTATION, CB_QUOTATION, CB_HIST_DATA, STK_HIST_DATA
from ldnode.loader.datasources.common_plans import plan_rename, plan_correct
from ldnode.loader.datasources.common_plans.limit_exprs import (
    cb_lower_limit_expr, 
    cb_upper_limit_expr,
)
from ldnode.loader.datasources.sina import SinaWithTimeoutPatch
from .ytm_plan import plan_ytm
from .conv_prem_plan import plan_conv_prem
from .history_plan import plan_history
from .option_plan import plan_option


class CBQuotationsDataSource(DataSource):
    def load(self, key: str) -> list[SyncResult]:
        """ 生成可转债实时截面数据 """
        try:
            if not CB_PRE_QUOTATION.get_local_path().exists() or not CB_HIST_DATA.get_local_path().exists() or not STK_HIST_DATA.get_local_path().exists():
                return [SyncResult(key, SyncStatus.FAILED, "数据源不存在")]
            # logger.loader.info(f"开始生成可转债实时截面数据")
            quotation = SinaWithTimeoutPatch()
            today = datetime.now().date()
            start_at = time.time()
            # 读取股票盘前数据
            pre_quotations = pl.read_parquet(CB_PRE_QUOTATION.get_local_path())
            # 下载行情数据
            download_at = time.time()
            ### 计算需要下载的标的代码列表
            codes = [
                *map(
                    lambda code: code.split('.')[1].lower() + code.split('.')[0], 
                    pre_quotations['code'],
                ),
                *map(
                    lambda code: code.split('.')[1].lower() + code.split('.')[0], 
                    pre_quotations['code_stk'],
                ),
            ]
            quotation_data = quotation.stocks(codes, prefix=True)
            # logger.loader.info(f"下载行情数据耗时: {humanize.naturaldelta(time.time() - download_at)}")
            # 计算可转债实时截面数据
            calculate_at = time.time()
            _quotations = (
                pl.DataFrame(quotation_data).transpose(include_header=True, header_name='code').lazy()
                .unnest('column_0')
                .pipe(plan_rename)
                .pipe(plan_correct)
            )
            quotations = (
                pre_quotations.lazy()
                .join(
                    _quotations,
                    left_on='code',
                    right_on='code',
                    how='left'
                )
                .drop('name_stk')
                .join(
                    _quotations,
                    left_on='code_stk',
                    right_on='code',
                    how='left',
                    suffix='_stk',
                )
                .pipe(plan_ytm, today)
                # 排除停牌
                .filter(
                    ~((col("open") == 0) & (col("bid1") == 0) & (col("ask1") == 0))
                )
                .with_columns(
                    (col('close') / col('pre_close') - 1).alias('pct_chg'),  # 转债涨跌幅
                    (col('close_stk') / col('pre_close_stk') - 1).alias('pct_chg_stk'),  # 正股涨跌幅
                    (col("amount") / 10000).alias("amount"),  # 成交额（单位：万）
                    (col("vol") / 10).alias("vol"),  # 成交量（单位：手）
                    (col("amount_stk") / 10000).alias("amount_stk"),  # 成交额（单位：万）
                    (col("vol_stk") / 100).alias("vol_stk"),  # 成交量（单位：手）
                    (100 / col("conv_price") * col("close_stk")).alias("conv_value"),  # 转股价值
                )
                .with_columns(
                    (col("close") / col("conv_value") - 1).alias("conv_prem"),  # 转股溢价率
                    (col("close") / col("pure_value") - 1).alias("bond_prem"),
                )
                .with_columns(  
                    # 纯债溢价率
                    (col("close") + col("conv_prem") * 100).alias("dblow"),  # 双低值
                    (col("pe_ttm") / col("pre_close_stk") * col("close_stk")).alias("pe_ttm"),
                    (col("pb") / col("pre_close_stk") * col("close_stk")).alias("pb"),
                    (col("ps_ttm") / col("pre_close_stk") * col("close_stk")).alias("ps_ttm"),
                    (col("total_mv") / col("pre_close_stk") * col("close_stk") / 1e4).alias("total_mv"),  # 总市值（单位：亿元）
                    (col("circ_mv") / col("pre_close_stk") * col("close_stk") / 1e4).alias("circ_mv"), # 流通市值（单位：亿元）
                    (col("remain_size") / 100 * col("close")).alias("remain_cap")  # 剩余市值
                )
                .with_columns(
                    (col("amount") / col("remain_cap") / 10000).alias("turnover"), # 换手率（注意：amount已被修改）
                    (col("remain_cap") / col("total_mv")).alias("cap_mv_rate"),  # 市占比
                    cb_upper_limit_expr.alias('upper_limit'), # 涨停价格
                    cb_lower_limit_expr.alias('lower_limit') # 跌停价格
                )
                .with_columns(
                    # 计算涨跌停标志
                    pl.when(col('close') == col('upper_limit')).then(pl.lit(1)).otherwise(
                        pl.when(col('close') == col('lower_limit')).then(pl.lit(-1)).otherwise(pl.lit(0))
                    ).cast(pl.Int8).alias('limit')
                )
                .pipe(plan_conv_prem)
                .pipe(plan_history)
                .pipe(plan_option)
            ).collect()
            # 对齐 S3 数据范式：S3 上的 cb_quotations.pq 使用 date（Datetime）而非 trade_date（Date），
            # 内部计算管道（plan_history 等）依赖 trade_date 列名，因此在输出前重命名回 date
            if 'trade_date' in quotations.columns:
                quotations = quotations.rename({'trade_date': 'date'})
            # 保存可转债实时截面数据
            quotations.write_parquet(CB_QUOTATION.get_local_path())
            # logger.loader.info(f"生成可转债实时截面数据总耗时: {humanize.naturaldelta(time.time() - start_at)}")
            return [SyncResult(key, SyncStatus.SUCCESS, today.strftime("%Y-%m-%d"))]
        except Exception as e:
            logger.loader.error(f"生成可转债实时截面数据失败: {e}")
            return [SyncResult(key, SyncStatus.FAILED, str(e))]



from typing import Optional, Iterator
from pathlib import Path

import humanize

from ldnode import logger
from ldnode.loader.schemas.enumerations import SyncStatus
from ldnode.loader.schemas.models import DataSource, SyncMetadata, SyncResult
from ldnode.loader.persistence import LightCOSClient


class COSDataSource(DataSource):
    """ COS 数据源 """
    def __init__(self, cos_client: LightCOSClient, metadata: SyncMetadata, local_path: Path):
        self.cos_client = cos_client
        self.local_path = local_path
        # 确保本地数据目录存在
        self.local_path.mkdir(parents=True, exist_ok=True)
        # 存储文件ETag的元数据文件
        self.metadata = metadata
        
    def _check_file_changed(self, filename: str) -> tuple[bool, Optional[tuple[str, str]]]:
        """
            检查文件是否发生变化
            @param filename: 文件名
            @return: 最新的ETAG
        """
        # 获取最新的ETAG
        if latest_tag := self.cos_client.get_file_etag(filename):
            # 获取当前的ETAG
            current_tag = self.metadata.get_file_etag(filename)
            return current_tag != latest_tag, (current_tag, latest_tag)
        else:
            return False, None
        
    def _log_bytes_transferred(self, operation: str, filename: str, bytes_transferred: int) -> None:
        """
            记录流量统计
            @param operation: 操作
            @param filename: 文件名
            @param bytes_transferred: 下载的字节数
        """
        logger.loader.info(f"流量统计 - {operation} {filename}: 共下载 {humanize.naturalsize(bytes_transferred)}({bytes_transferred:,} 字节)")
        
    def _download_file(self, filename: str, latest_etag: str, local_file_path: Path) -> bool:
        """
            下载文件
            @param filename: 文件名
            @param local_file_path: 本地文件路径
            @return: 是否下载成功
        """
        logger.loader.info(f"开始全量同步文件: {filename}")
        try:
            # 重置流量计数器
            self.cos_client.reset_bytes_transferred()
            # 下载文件
            self.cos_client.download_file(filename, local_file_path)
            # 更新ETag
            self.metadata.set_file_etag(filename, latest_etag)
            # 显示总流量消耗
            self._log_bytes_transferred("总计", filename, self.cos_client.get_bytes_transferred())
            return True
        except Exception as e:
            logger.loader.error(f"全量同步文件失败 {filename}: {e}")
            return False
        
    def _get_file_path(self, filename: str) -> Path:
        """
            获取文件路径
            @param filename: 文件名
            @return: 文件路径
        """
        return self.local_path / filename
    
    def _walk_files(self, remote_path: str) -> Iterator[str]:
        """
            遍历文件
            @param remote_path: 远程路径
            @return: 文件列表
        """
        for filename in self.cos_client.list_objects(remote_path):
            yield filename
        
    def load(self, key: str) -> list[SyncResult]:
        results: list[SyncResult] = []
        for filename in self._walk_files(key):
            try:
                # 获取文件路径
                local_file_path = self._get_file_path(filename)
                # 检查文件是否发生变化
                file_changed, (current_tag, latest_tag) = self._check_file_changed(filename)
                # 如果文件不存在, 则下载文件
                if not local_file_path.exists() or file_changed:
                    # 文件发生变化, 下载文件
                    logger.loader.info(f"文件 {filename} 发生变化，执行全量同步, ETag: {current_tag} => {latest_tag}")
                    self._download_file(filename, latest_tag, local_file_path)
                    results.append(SyncResult(filename, SyncStatus.SUCCESS, latest_tag))
                else:
                    results.append(SyncResult(filename, SyncStatus.NO_CHANGE, latest_tag))
            except Exception as e:
                logger.loader.error(f"下载文件 {filename} 失败: {e}")
                results.append(SyncResult(filename, SyncStatus.FAILED, str(e)))
        return results



import polars as pl

from ldnode.loader.files import INDEX_DATA
from ldnode.calculator.schemas import BenchmarkType


def get_index_data_lf(benchmark: BenchmarkType) -> pl.LazyFrame:
    """ 获取指数数据 """
    match benchmark:
        case BenchmarkType.EQUALWEIGHT:
            benchmark_column = 'JSL'
        case BenchmarkType.GZ2000:
            benchmark_column = '399303.SZ'
        case BenchmarkType.HS300:
            benchmark_column = '000300.SH'
        case BenchmarkType.ZZ1000:
            benchmark_column = '000852.SH'
        case BenchmarkType.STOCK:
            benchmark_column = 'JSL' # 可转债基准为正股尚未实现
        case _:
            benchmark_column = str(benchmark) # 自选基准，直接用原始字符串做列名
    try:
        return (
            INDEX_DATA.output_lazyframe()
            .select(
                pl.col('trade_date').cast(pl.Date),
                pl.col(benchmark_column).alias('time_return_benchmark') # 基准每周期收益率
            )
        )
    except Exception as e:
        raise ValueError(f"获取基准数据失败: {e}")



from io import BytesIO

from ldnode.loader.persistence import get_clickhouse_client
from ldnode.calculator.operators import cb_grid_backtest
from .models import CBGridBacktestParameters, CBGridBacktestResponse, TaskResponse
from .utils import wrap_handler, response, error_response


@wrap_handler("转债网格回测")
def handle_cb_grid_backtest(parameters: CBGridBacktestParameters) -> TaskResponse[CBGridBacktestResponse]:
    """ 响应转债网格回测任务 """
    # 获取Clickhouse客户端
    clickhouse_client = get_clickhouse_client()
    # 执行查询
    df = clickhouse_client.query_dataframe(
        """
            select open, toTimeZone(datetime, 'Asia/Shanghai') as datetime, close from cb_data_1m
            where symbol=%(code)s
            and (datetime between %(start_date)s and %(end_date)s)
            order by datetime
        """,
        {
            'code': parameters['strategy_config']['code'], 
            'start_date': parameters['strategy_config']['start_date'], 
            'end_date': parameters['strategy_config']['end_date']
        }
    )
    if df.empty:
        return error_response("所选时间段内没有发生任何数据，请重新选择时间段或标的")
    df = cb_grid_backtest(
        df=df,
        strategy_config=parameters['strategy_config']
    )
    evaluate_dict = {
        'start_time': df.index[0].to_pydatetime().strftime('%Y-%m-%d %H:%M'), 
        'start_pos': df.iat[0, df.columns.get_loc('pos')] if df.iat[0, df.columns.get_loc('pos')] else 0,
        'start_pos_val': df.iat[0, df.columns.get_loc('pos_val')], 
        'start_cash': df.iat[0, df.columns.get_loc('cash')],
        'end_time': df.index[-1].to_pydatetime().strftime('%Y-%m-%d %H:%M'), 
        'end_pos': df.iat[-1, df.columns.get_loc('pos')], 
        'end_pos_val': df.iat[-1, df.columns.get_loc('pos_val')], 
        'end_cash': df.iat[-1, df.columns.get_loc('cash')], 
        'end_asset': df.iat[-1, df.columns.get_loc('asset')],
        'times': f'{(df.index[-1] - df.index[0]).days}天', 
        'cash_mean': df.cash.mean(), 
        'cash_max': df.cash.min(), 
        'grid_return_acc': df.iat[-1, df.columns.get_loc('grid_return_acc')]
    }
    df = df.reset_index()[
        ['datetime','open','close', 'grid_index', 'pos','pos_val','cash','grid_return','grid_return_acc','asset']
    ]
    pq_buffer = BytesIO()
    df.to_parquet(pq_buffer, index=True)
    return response(data_type="grid_backtest_result", data={
        "evaluation": evaluate_dict,
        "history_position": pq_buffer.getvalue(),
    })
